import sys
from dan.core.cache import Cache

from dan.core.errors import InvalidConfiguration
from dan.core.settings import Settings
from dan.cxx.toolchain import Toolchain
from dan.cxx.detect import get_toolchains

target_toolchain: Toolchain = None
"""The target toolchain.
"""

host_toolchain: Toolchain = None
"""The host toolchain.
"""

class __LazyContext(sys.__class__):
    """Base class for the cxx module.

    It overloads some context dependent properties exposed by this module, eg.: target and host toolchains.
    """
    @property
    def target_toolchain(__) -> Toolchain:
        from dan.core.include import context
        return context.get('cxx_target_toolchain')

    @property
    def host_toolchain(__) -> Toolchain:
        from dan.core.include import context
        return context.get('cxx_host_toolchain')
    

sys.modules[__name__].__class__ = __LazyContext


auto_fpic = True

def get_default_toolchain(data = None):
    data = data or get_toolchains()
    return data['default']


def init_toolchains(name: str = None, settings: Settings = None):
    data = get_toolchains()
    if name is None or name == 'default':
        name = get_default_toolchain(data)

    toolchain_data = data['toolchains'][name]

    tc_type = toolchain_data['type']
    match tc_type:
        case 'gcc' | 'clang':
            from .unix_toolchain import UnixToolchain
            tc_type = UnixToolchain
        case 'msvc':
            from .msvc_toolchain import MSVCToolchain
            tc_type = MSVCToolchain
        case _:
            raise InvalidConfiguration(f'Unhandeld toolchain type: {tc_type}')
    target_settings = settings.target
    cache = Cache.get('dan').data
    if not 'toolchains' in cache:
        cache['toolchains'] = {
            'host': dict(),
            'target': dict(),
        }
    target_toolchain = tc_type(toolchain_data, data['tools'], settings=target_settings, cache=cache['toolchains']['target'])
    target_toolchain.init()
    if target_toolchain.is_host:
        host_toolchain = target_toolchain
    else:
        import logging
        logging.warning(f'Cross compilation is currently not tested !')
        host_toolchain = None

    from dan.core.include import context
    context.set('cxx_target_toolchain', target_toolchain)
    context.set('cxx_host_toolchain', host_toolchain)


from .targets import Executable, Library, LibraryType, Module
from .targets import CXXObjectsTarget as Objects



import os
import pathlib

class Path(type(pathlib.Path())):

    @property
    def modification_time(self):
        return self.stat().st_mtime

    def younger_than(self, other):
        time = other if isinstance(other, float) else other.modification_time
        assert isinstance(time, float)
        return self.modification_time > time

    def older_than(self, other):
        time = other if isinstance(other, float) else other.modification_time
        assert isinstance(time, float)
        return self.modification_time < time

    def utime(self, *args, **kw_args):
        os.utime(self, *args, **kw_args)

    @property
    def is_empty(self):
        if os.path.exists(self) and not os.path.isfile(self):
    
            # Checking if the directory is empty or not
            if not os.listdir(self):
                return True
            else:
                return False
        else:
            return False



from functools import cached_property
import functools
from dan.core.register import MakefileRegister
from dan.core.pathlib import Path
from typing import Any, Callable, Iterable, Union, TypeAlias
import inspect

from dan.core import asyncio, aiofiles, utils, diagnostics as diags
from dan.core.cache import Cache
from dan.core.requirements import load_requirements
from dan.core.settings import InstallMode, InstallSettings, safe_load
from dan.core.version import Version
from dan.logging import Logging


class Dependencies:

    def __init__(self, parent: 'Target', deps: Iterable = list()):
        super().__init__()
        self.parent = parent
        self._content = list()
        for dep in deps:
            self.add(dep)

    @property
    def makefile(self):
        return self.parent.makefile

    def add(self, dependency):
        if dependency in self._content:
            return
        from dan.pkgconfig.package import RequiredPackage
        match dependency:
            case Target() | FileDependency():
                self._content.append(dependency)
            case type():
                assert issubclass(dependency, Target)
                self._content.append(self.makefile.find(dependency))
            case str():
                from dan.pkgconfig.package import Package
                for pkg in Package.all.values():
                    if pkg.name == dependency:
                        self._content.append(pkg)
                        break
                else:
                    if Path(self.parent.source_path / dependency).exists():
                        self._content.append(FileDependency(
                            self.parent.source_path / dependency))
                    else:
                        from dan.pkgconfig.package import parse_requirement
                        self._content.append(parse_requirement(dependency))
            case Path():
                dependency = FileDependency(
                    self.parent.source_path / dependency)
                self._content.append(dependency)
            case RequiredPackage():
                self._content.append(dependency)
            case _:
                raise RuntimeError(
                    f'Unhandled dependency {dependency} ({type(dependency)})')

    def update(self, dependencies):
        for dep in dependencies:
            self.add(dep)

    def __getattr__(self, attr):
        for item in self._content:
            if item.name == attr:
                return item

    def __iter__(self):
        return self._content.__iter__()

    @property
    def up_to_date(self):
        for item in self:
            if not item.up_to_date:
                return False
        return True

    @property
    def modification_time(self):
        t = 0.0
        for item in self:
            mt = item.modification_time
            if mt and mt > t:
                t = mt
        return t


TargetDependencyLike: TypeAlias = Union[list['Target'], 'Target']


PathImpl = type(Path())


class FileDependency(PathImpl):
    up_to_date = True

    def __init__(self, *args, **kwargs):
        super(PathImpl, self).__init__()

    @property
    def modification_time(self):
        return self.stat().st_mtime


class Option:
    def __init__(self, parent: 'Options', fullname: str, default, help: str = None) -> None:
        self.fullname = fullname
        self.name = fullname.split('.')[-1]
        self.__parent = parent
        self.__cache = parent._cache
        self.__default = default
        self.__value = self.__cache.get(self.name, default)
        self.__value_type = type(default)
        self.__help = help if help is not None else 'No description.'

    def reset(self):
        self.value = self.__default

    @property
    def parent(self):
        return self.__parent

    @property
    def cache(self):
        return self.__cache

    @property
    def type(self):
        return self.__value_type

    @property
    def default(self):
        return self.__default

    @property
    def help(self):
        return self.__help

    @property
    def value(self):
        return self.__value

    @value.setter
    def value(self, value):
        value = safe_load(self.fullname, value, self.__value_type)
        if self.__value != value:
            self.__value = value
            self.__cache[self.name] = value


class Options:
    def __init__(self, parent: 'Target', default: dict[str, Any] = dict()) -> None:
        self.__parent = parent
        cache = parent.cache
        if isinstance(parent.cache, dict):
            if not parent.name in cache:
                cache[parent.name] = dict()
            cache = cache[parent.name]
        else:
            cache = parent.cache.data
        if not 'options' in cache:
            cache['options'] = dict()
        self._cache = cache['options']
        self.__items: list[Option] = list()
        self.update(default)

    def add(self, name: str, default_value, help=None):
        opt = Option(self, f'{self.__parent.fullname}.{name}',
                     default_value, help=help)
        self.__items.append(opt)
        return opt

    def get(self, name: str):
        for o in self.__items:
            if name in {o.name, o.fullname}:
                return o

    def update(self, options: dict):
        for k, v in options.items():
            help = None
            match v:
                case dict():
                    help = v['help']
                    v = v['default']
                case tuple() | list() | set():
                    help = v[1]
                    v = v[0]
                case _:
                    pass
            if self[k]:
                self[k] = v
            else:
                self.add(k, v, help)
    
    @property
    def sha1(self):
        import hashlib
        sha1 = hashlib.sha1()
        for o in self.__items:
            sha1.update(o.fullname.encode() + str(o.value).encode())
        return sha1.hexdigest()

    def items(self):
        for o in self.__items:
            yield o.name, o.value

    def __getattr__(self, name):
        opt = self.get(name)
        if opt:
            return opt.value

    def __getitem__(self, name):
        opt = self.get(name)
        if opt:
            return opt.value

    def __iter__(self):
        return iter(self.__items)


class Target(Logging, MakefileRegister, internal=True):
    name: str = None
    fullname: str = None
    description: str = None,
    default: bool = True
    installed: bool = False
    output: Path = None
    options: dict[str, Any] = dict()

    dependencies: set[TargetDependencyLike] = set()
    preload_dependencies: set[TargetDependencyLike] = set()

    def __init__(self,
                 name: str = None,
                 parent: 'Target' = None,
                 version: str = None,
                 default: bool = None,
                 makefile=None) -> None:
        
        self.parent = parent
        self.__cache: dict = None

        if name is not None:
            self.name = name

        if self.name is None:
            self.name = self.__class__.__name__

        if default is not None:
            self.default = default

        if parent is not None:
            self.makefile = parent.makefile
            self.fullname = f'{parent.fullname}.{self.name}'

        if makefile:
            self.makefile = makefile

        if self.makefile is None:
            raise RuntimeError('Makefile not resolved')


        if self.fullname is None:
            self.fullname = f'{self.makefile.fullname}.{self.name}'

        self.options = Options(self, self.options)

        if version is not None:
            self._version = version

        if not hasattr(self, '_version'):
            self._version = self.makefile.version

        if self.description is None:
            self.description = self.makefile.description

        self.other_generated_files: set[Path] = set()
        self.dependencies = Dependencies(self, self.dependencies)
        self.preload_dependencies = Dependencies(
            self, self.preload_dependencies)

        super().__init__(self.fullname)

        self._output: Path = None

        if type(self).output != Target.output:
            # hack class-defined output
            #   transform it to classproperty for build_path resolution
            output = self.output
            type(self).output = utils.classproperty(lambda: self.build_path / output)

        self.diagnostics = diags.DiagnosticCollection()

    @property
    def output(self):
        if self._output is None:
            return None
        return self.build_path / self._output

    @property
    def routput(self):
        return self._output
    
    @property
    def version(self):
        version = self._version
        if isinstance(version, Option):
            version = version.value
        if isinstance(version, str):
            version = Version(version)
        return version

    @version.setter
    def version(self, value):
        self._version = value

    @output.setter
    def output(self, path):
        path = Path(path)
        if not path.is_absolute() and self.build_path in path.parents:
            raise RuntimeError(f'output must not be an absolute path within build directory')
        elif path.is_absolute() and self.build_path in path.parents:
            self._output = path.relative_to(self.build_path)
        else:
            self._output = path

    @property
    def is_requirement(self) -> bool:
        return self.makefile.name.endswith('requirements')

    @property
    def source_path(self) -> Path:
        return self.makefile.source_path

    @property
    def build_path(self) -> Path:
        return self.makefile.build_path
    
    @property
    def requires(self):
        from dan.pkgconfig.package import RequiredPackage
        return [dep for dep in self.dependencies if isinstance(dep, RequiredPackage)]

    @cached_property
    def fullname(self) -> str:
        return f'{self.makefile.fullname}.{self.name}'

    @property
    def cache(self) -> dict:
        if not self.__cache:
            name = self.fullname.removeprefix(self.makefile.fullname + '.')
            if not name in self.makefile.cache.data:
                self.makefile.cache.data[name] = dict()
            self.__cache = self.makefile.cache.data[name]
        return self.__cache

    async def __load_unresolved_dependencies(self):
        if len(self.requires) > 0:
            self.dependencies.update(await load_requirements(self.requires, makefile=self.makefile, logger=self))

    @asyncio.cached
    async def preload(self):
        self.debug('preloading...')

        async with asyncio.TaskGroup(f'building {self.name}\'s preload dependencies') as group:
            group.create_task(self.__load_unresolved_dependencies())
            for dep in self.preload_dependencies:
                group.create_task(dep.build())

        async with asyncio.TaskGroup(f'preloading {self.name}\'s target dependencies') as group:
            for dep in self.target_dependencies:
                group.create_task(dep.preload())

        res = self.__preload__()
        if inspect.iscoroutine(res):
            res = await res
        return res

    @asyncio.cached
    async def initialize(self):
        await self.preload()
        self.debug('initializing...')

        if isinstance(self.version, Option):
            self.version = self.version.value

        async with asyncio.TaskGroup(f'initializing {self.name}\'s target dependencies') as group:
            for dep in self.target_dependencies:
                group.create_task(dep.initialize())

        res = self.__initialize__()
        if inspect.iscoroutine(res):
            res = await res
        return res

    @property
    def modification_time(self):
        output = self.build_path / f'{self.name}.stamp' if self.output is None else self.output  
        return output.stat().st_mtime if output.exists() else 0.0

    @cached_property
    def up_to_date(self):
        output = self.build_path / f'{self.name}.stamp' if self.output is None else self.output
        if output and not output.exists():
            return False
        elif not self.dependencies.up_to_date:
            return False
        elif self.dependencies.modification_time > self.modification_time:
            return False
        elif 'options_sha1' in self.cache and self.cache['options_sha1'] != self.options.sha1:
            return False
        return True

    async def _build_dependencies(self):
        async with asyncio.TaskGroup(f'building {self.name}\'s target dependencies') as group:
            for dep in self.target_dependencies:
                group.create_task(dep.build())

    @asyncio.cached
    async def build(self):
        await self.initialize()

        await self._build_dependencies()

        result = self.__prebuild__()
        if inspect.iscoroutine(result):
            await result

        if self.up_to_date:
            self.info('up to date !')
            return
        elif self.output is not None and self.output.exists():
            self.info('outdated !')

        with utils.chdir(self.build_path):
            self.info('building...')
            if diags.enabled:
                self.diagnostics.clear()
            result = self.__build__()
            if inspect.iscoroutine(result):
                result = await result
            if self.output is None:
                (self.build_path / f'{self.name}.stamp').touch()
            self.cache['options_sha1'] = self.options.sha1
            return result

    @property
    def target_dependencies(self):
        return [t for t in {*self.dependencies, *self.preload_dependencies} if isinstance(t, Target)]

    @property
    def file_dependencies(self):
        return [t for t in self.dependencies if isinstance(t, FileDependency)]

    @asyncio.cached
    async def clean(self):
        await self.initialize()
        async with asyncio.TaskGroup(f'cleaning {self.name} outputs') as group:
            output = self.build_path / f'{self.name}.stamp' if self.output is None else self.output
            if output and output.exists():
                self.info('cleaning...')
                if output.is_dir():
                    group.create_task(aiofiles.rmtree(output, force=True))
                else:
                    group.create_task(aiofiles.os.remove(output))
            for f in self.other_generated_files:
                if f.exists():
                    group.create_task(aiofiles.os.remove(f))
            res = self.__clean__()
            if inspect.iscoroutine(res):
                group.create_task(res)

    @asyncio.cached(unique = True)
    async def install(self, settings: InstallSettings, mode: InstallMode):
        await self.build()

        self.debug('installing %s to %s', self.name, settings.destination)

        installed_files = list()
        if mode == InstallMode.dev:
            if len(self.utils) > 0:
                lines = list()
                for fn in self.utils:
                    tmp = inspect.getsourcelines(fn)[0]
                    tmp[0] = f'\n\n@self.utility\n'
                    lines.extend(tmp)
                filepath = settings.libraries_destination / \
                    'dan' / f'{self.name}.py'
                filepath.parent.mkdir(exist_ok=True, parents=True)
                async with aiofiles.open(filepath, 'w') as f:
                    await f.writelines(lines)
                    installed_files.append(filepath)
        return installed_files

    def get_dependency(self, dep: str | type, recursive=True) -> TargetDependencyLike:
        """Search for dependency"""
        if isinstance(dep, str):
            def check(d): return d.name == dep
        else:
            def check(d): return isinstance(d, dep)
        for dependency in self.dependencies:
            if check(dependency):
                return dependency
        for dependency in self.preload_dependencies:
            if check(dependency):
                return dependency
        if recursive:
            # not found... look for dependencies' dependencies
            for target in self.target_dependencies:
                dependency = target.get_dependency(dep)
                if dependency is not None:
                    return dependency

    async def __preload__(self):
        ...

    async def __initialize__(self):
        ...

    async def __prebuild__(self):
        ...

    async def __build__(self):
        ...

    async def __install__(self):
        ...

    async def __clean__(self):
        ...

    @utils.classproperty
    def utils(cls) -> list:
        utils_name = f'_{cls.__name__}_utils__'
        if not hasattr(cls, utils_name):
            setattr(cls, utils_name, list())
        return getattr(cls, utils_name)

    @classmethod
    def utility(cls, fn: Callable):
        cls.utils.append(fn)
        name = fn.__name__
        makefile = cls.get_static_makefile()
        if makefile is not None:
            inst = makefile.find(cls)
            fn = functools.partial(fn, inst)
        setattr(cls, name, fn)
        return fn




import json
from dan.core.pathlib import Path
import subprocess


class CompileCommands:
    def __init__(self, path) -> None:
        self.cc_path: Path = path / 'compile_commands.json'
        if self.cc_path.exists():
            with open(self.cc_path, 'r') as cc_f:
                try:
                    self.data = json.load(cc_f)
                except json.JSONDecodeError:
                    self.data = list()
        else:
            self.data = list()
            self.cc_path.parent.mkdir(parents=True, exist_ok=True)

    def clear(self):
        with open(self.cc_path, 'w'):
            pass

    def update(self):
        with open(self.cc_path, 'w') as cc_f:
            json.dump(self.data, cc_f)

    def get(self, file: Path):
        fname = file.name
        for entry in self.data:
            if entry['file'] == fname:
                return entry
        return None

    def insert(self, file: Path, build_path: Path, content: list[str] | str):
        entry = self.get(file)
        if isinstance(content, list):
            content = subprocess.list2cmdline(content)

        if entry:
            entry['command'] = content
        else:
            self.data.append({
                'file': str(file),
                'directory': str(build_path),
                'command': content
            })



from dan.core.pathlib import Path
from dan.core.settings import InstallMode
from tests import PyMakeBaseTest


class CXXSimpleLibTest(PyMakeBaseTest):
    def __init__(self, methodName: str = None) -> None:
        super().__init__('cxx/libraries', methodName)

    async def test_build(self):
        target_name = 'simplelib'

        ########################################
        async with self.section("base build", clean=True) as make:
            target = make.root.find(target_name)
            await target.initialize()
            self.assertFalse(target.output.exists())
            await target.build()
            self.assertTrue(target.output.exists())
            self.modified_at = target.output.modification_time

        # ########################################
        # async with self.section("no-modification => no-rebuild") as make:
        #     target = make.root.find(target_name)
        #     await target.build()
        #     self.assertTrue(target.output.exists())
        #     self.assertEqual(self.modified_at, target.output.modification_time,
        #                     "no modifications should NOT trigger a re-build")

        #     # update source
        #     src: Path = target.source_path / list(target.sources)[0]
        #     src.utime()

        # async with self.section("source modification => rebuild") as make:
        #     target = make.root.find(target_name)
        #     await target.build()
        #     self.assertTrue(target.output.younger_than(self.modified_at),
        #                     "a source modification should trigger a re-build")
        #     self.modified_at = target.output.modification_time


    # async def test_install(self):
    #     target_name = 'simplelib'

    #     ########################################
    #     async with self.section("base build",
    #                             options=[
    #                                 'root.library_type=shared'
    #                             ],
    #                             settings=[
    #                                 f'install.destination={self.build_path}/dist'
    #                             ],
    #                             clean=True) as make:
    #         self.assertEqual(make.settings.install.runtime_destination, self.build_path / 'dist/bin')
    #         target = make.root.find(target_name)
    #         await target.initialize()
    #         self.assertFalse(target.output.exists())
    #         await target.build()
    #         self.assertTrue(target.output.exists())
    #         modified_at = target.output.modification_time


    #     ########################################
    #     async with self.section("install", init=False) as make:
    #         await make.install(InstallMode.user)
    #         target = make.root.find(target_name)
    #         # should be re-linked with new rpath
    #         self.assertTrue(target.output.younger_than(modified_at))
    #         self.assertTrue(
    #             (self.build_path / f'dist/bin/{target_name}').exists())




from dan.core.pathlib import Path
from dan.core import asyncio
from dan.core.target import Target, TargetDependencyLike
from typing import Callable
import inspect


class generator:
    def __init__(self, output: str, dependencies: TargetDependencyLike = None, options: dict = None):
        self.output = Path(output)
        self.dependencies = list() if dependencies is None else dependencies
        self.options = dict() if options is None else options

    def __call__(self, fn: Callable):
        class Generator(Target):
            name = self.output.stem
            output = self.output
            dependencies = set(self.dependencies)
            options = self.options

            def __build__(self):
                arg_spec = inspect.getfullargspec(fn)
                if 'self' in arg_spec.args:
                    return fn(self)
                elif not arg_spec.args:
                    return fn()
                else:
                    raise RuntimeError(
                        "Only 'self' is allowed as Generator argument")

        # hack the module location (used for Makefile's Targets resolution)
        Generator.__module__ = fn.__module__
        return Generator



import json
import aiohttp

from dan.core.version import Version
from dan.core import asyncio
from dan.src.tar import TarSources

class GitHubReleaseSources(TarSources, internal=True):

    user : str
    project : str
    use_tags = False

    @property
    def url(self):
        return self._url

    def __init__(self, *args, **kwargs):
        if self.name is None:
            self.name = f'{self.user}-{self.project}-sources'
        super().__init__(*args, **kwargs)

    async def __initialize__(self):
        await super().__initialize__()

        self._url = self.cache.get('url')
        if self._url is not None and self.version is not None:
            prev_version = self.cache.get('version')
            if prev_version is not None and prev_version != self.version:
                self._url = None
            
        if self._url is None:
            avail_versions = await self.available_versions()
            if self.version is None:
                self.version = sorted(avail_versions, reverse=True)[0]
            version = self.version if isinstance(self.version, Version) else Version(self.version)
            release = avail_versions[version]
            if self.use_tags:
                if 'tarball_url' in release:
                    self._url = release['tarball_url']
            else:
                assets = release['assets']
                # prefer assets over tarball, asset might contain submodules while tarballs will not
                for asset in assets:
                    if asset['content_type'] in ('application/zip', 'application/gzip'):
                        self._url = asset['browser_download_url']
                        break
                else:
                    if 'tarball_url' in release:
                        self._url = release['tarball_url']
            if self._url is None:
                raise RuntimeError(f'cannot resolve url of {self.name}')
            self.cache['version'] = self.version
            self.cache['url'] = self._url

    @asyncio.cached
    async def available_versions(self) -> dict[Version, dict]:
        self.info('fetching github releases')


        api_token = None

        from dan.io.repositories import _get_settings
        settings = _get_settings()
        if settings.github.api_token is not None:
            api_token = settings.github.api_token

        if self.use_tags:
            url = f'https://api.github.com/repos/{self.user}/{self.project}/tags'
            async with aiohttp.ClientSession() as session:
                if api_token is not None:
                    session.headers['Authorization'] = f'Bearer {api_token}'
                async with session.get(url) as resp:
                    data = await resp.read()
                    if resp.status != 200:
                        raise RuntimeError(f'unable to fetch {url}: {data.decode()}')
                    releases = json.loads(data)
                    return {Version(release['name']): release for release in releases}
        else:
            url = f'https://api.github.com/repos/{self.user}/{self.project}/releases'
            async with aiohttp.ClientSession() as session:
                if api_token is not None:
                    session.headers['Authorization'] = f'Bearer {api_token}'
                async with session.get(url) as resp:
                    data = await resp.read()
                    if resp.status != 200:
                        raise RuntimeError(f'unable to fetch {url}: {data.decode()}')
                    releases = json.loads(data)
                    return {Version(release['tag_name']): release for release in releases}





from tests import PyMakeBaseTest


class CXXCatch2Test(PyMakeBaseTest):
    def __init__(self, methodName: str = None) -> None:
        super().__init__('cxx/src/catch2', methodName)

    async def test_config(self):

        ########################################
        async with self.section("base build", clean=True) as make:
            config = make.root.requirements.find('catch2-config')
            await config.build()
            self.assertTrue(config.output.exists())
            modified_at = config.output.modification_time

            self.assertTrue(config.up_to_date)
            ########################################
            counter = config.options.get('counter')
            counter.value = not counter.value
            self.assertFalse(config.up_to_date)

        async with self.section("option modification => rebuild") as make:
            config = make.root.requirements.find('catch2-config')
            await config.build()
            self.assertTrue(config.output.younger_than(modified_at),
                            "an option change should trigger a re-build")




import asyncio
import fnmatch
import os
import contextlib

from dan.cli import click
from dan.core.requirements import parse_package
from dan.make import Make

_make : Make = None
async def get_make(toolchain='default', quiet=True):
    global _make
    if _make is None:
        from dan.cxx.detect import get_dan_path
        source_path = get_dan_path() / 'deps'
        source_path.mkdir(exist_ok=True, parents=True)
        os.chdir(source_path)
        (source_path / 'dan-build.py').touch()
        make = Make(source_path / 'build', quiet=quiet)
        make.config.source_path = str(source_path)
        make.config.build_path = str(source_path / 'build')
        make.config.toolchain = toolchain
        await make._config.save()
        await make.initialize()
        _make = make
    return _make


_repositories = None
async def get_repositories():
    global _repositories
    if _repositories is None:
        from dan.io.repositories import get_all_repo_instances
        await get_make()
        _repositories = get_all_repo_instances()
        async with asyncio.TaskGroup() as g:
            for repo in _repositories:
                g.create_task(repo.build())
    return _repositories

async def get_repository(name = None):
    from dan.io.repositories import get_repo_instance
    await get_make()
    repo = get_repo_instance(name)
    await repo.build()
    return repo


@contextlib.asynccontextmanager
async def make_context(toolchain='default', quiet=True):
    make = await get_make(toolchain, quiet=quiet)
    with make.context:
        yield make


@click.group()
def cli():
    pass

@cli.group()
def ls():
    """Inspect stuff"""
    pass


@ls.command()
async def repositories():
    """List available repositories"""
    async with make_context():
        repos = await get_repositories()
        for repo in repos:
            click.echo(repo.name)

@ls.command()
async def libraries():
    """List available libraries"""
    async with make_context():
        repos = await get_repositories()
        for repo in repos:
            for name, lib in repo.installed.items():
                click.echo(f'{name} = {lib.version}')

async def get_library(library_spec):
    package, library, repository = parse_package(library_spec)
    repo = await get_repository(repository)
    if repo is None:
        raise RuntimeError(f'cannot find repository {repository}')

    lib = repo.find(library, package)
    if lib is None:
        if repository is None:
            repository = repo.name
        if package is None:
            package = library
        raise RuntimeError(f'cannot find {package}:{library}@{repository}')

    return lib

@ls.command()
@click.argument('LIBRARY')
async def versions(library: str):
    """Get LIBRARY's available versions"""
    async with make_context():
        lib = await get_library(library)
        
        from dan.src.github import GitHubReleaseSources
        
        sources: GitHubReleaseSources = lib.get_dependency(GitHubReleaseSources)
        available_versions = await sources.available_versions()
        available_versions = sorted(available_versions.keys())
        for v in available_versions:
            if v == lib.version:
                click.echo(f' - {v} (default)')
            else:
                click.echo(f' - {v}')

@cli.command()
@click.argument('NAME')
async def search(name):
    """Search for NAME in repositories"""
    async with make_context():
        name = f'*{name}*'
        repos = await get_repositories()
        for repo in repos:
            installed = repo.installed
            for libname, lib in installed.items():
                if fnmatch.fnmatch(libname, name):
                    click.echo(f'{libname} = {lib.version}')

@cli.command()
@click.option('--toolchain', '-t', type=click.ToolchainParamType(), default='default')
@click.argument('PACKAGE_SPEC')
@click.argument('VERSION', required=False)
async def install(toolchain, package_spec, version):
    """Intall given PACKAGE_SPEC"""
    from dan.io.package import PackageBuild

    async with make_context(toolchain, quiet=False) as make:
        package, name, repository = parse_package(package_spec)
        pkg = PackageBuild(name, version, package, repository, makefile=make.root)
        await pkg.initialize()
        if pkg.up_to_date:
            click.echo(f'Package {package_spec} already installed at version {pkg.version}')
        else:
            await pkg.build()
            click.echo(f'Package {package_spec} installed successfully at version {pkg.version}')
            

def main():
    import sys
    try:
        cli(auto_envvar_prefix='DAN')
    except Exception as err:
        click.logger.error(str(err))
        _ex_type, _ex, tb = sys.exc_info()
        import traceback
        click.logger.debug(' '.join(traceback.format_tb(tb)))
        try:
            # wait asyncio loop to terminate
            asyncio.get_running_loop().run_until_complete()
        except Exception:
            pass
        return -1

if __name__ == '__main__':
    main()



from contextlib import contextmanager
import importlib.util
import os

from dan.core.asyncio import sync_wait
from dan.core.makefile import MakeFile
from dan.core.pathlib import Path
from dan.core.requirements import load_requirements
from dan.core.target import Target
from dan.logging import Logging
from dan.pkgconfig.package import parse_requirement


class TargetNotFound(RuntimeError):
    def __init__(self, name) -> None:
        super().__init__(f'package {name} not found')


def requires(*requirements) -> list[Target]:
    ''' Requirement lookup

    1. Searches for a target exported by a previously included makefile
    2. Searches for pkg-config library

    :param names: One (or more) requirement(s).
    :return: The list of found targets.
    '''
    # return [parse_requirement(req) for req in requirements]
    global context
    requirements = [parse_requirement(req) for req in requirements]
    sync_wait(load_requirements(requirements, makefile=context.current, install=False))
    return requirements


class Context(Logging):
    def __init__(self) -> None:
        self.__root: MakeFile = None
        self.__current: MakeFile = None
        self.__all_makefiles: set[MakeFile] = set()
        self.imported_makefiles: dict[Path, MakeFile] = dict()
        self.__prev_ctx: Context = None
        self.__attributes = dict()
        super().__init__('context')

    @property
    def root(self):
        return self.__root

    @property
    def current(self):
        return self.__current

    @property
    def all_makefiles(self) -> set[MakeFile]:
        return self.__all_makefiles

    @current.setter
    def current(self, current: MakeFile):
        if self.__root is None:
            self.__root = current
            self.__current = current
        else:
            current.parent = self.__current
            self.__current = current

        self.__all_makefiles.add(current)

    def up(self):
        if self.__current != self.__root:
            self.__current = self.__current.parent

    def get(self, name, default=None):
        if name in self.__attributes:
            return self.__attributes[name]
        if default is not None:
            self.__attributes[name] = default
            return default

    def set(self, name, value):
        self.__attributes[name] = value

    def __enter__(self):
        global context
        self.__prev_ctx = context
        context = self
        return self

    def __exit__(self, *exc):
        global context
        context = self.__prev_ctx
        self.__prev_ctx = None


context: Context = Context()

class MakeFileError(RuntimeError):
    def __init__(self, path) -> None:
        self.path = Path(path)
        super().__init__(f'failed to load {self.path}')


def _init_makefile(module, name: str = 'root', build_path: Path = None, requirements: MakeFile = None, parent: MakeFile = None):
    global context
    source_path = Path(module.__file__).parent
    if parent is None and context.current is not None:
        parent = context.current

    if build_path is None:
        assert parent is not None
        build_path = build_path or parent.build_path / name

    module.__class__ = MakeFile
    context.current = module
    module._setup(
        name,
        source_path,
        build_path,
        requirements,
        parent)

def load_makefile(module_path: Path, name: str = None, module_name: str = None, build_path: Path = None, requirements: MakeFile = None, parent: MakeFile = None) -> MakeFile:
    name = name or module_path.stem
    module_name = module_name or name
    if module_path in context.imported_makefiles:
        return context.imported_makefiles[module_path]
    spec = importlib.util.spec_from_file_location(
        module_name, module_path)
    module = importlib.util.module_from_spec(spec)
    context.imported_makefiles[module_path] = module
    _init_makefile(module, name, build_path, requirements, parent)
    try:
        spec.loader.exec_module(module)
    except Exception as err:
        context.error('makefile error while loading \'%s\': %s', module_path, err)
        raise MakeFileError(module_path) from err
    context.up()
    return module


def include_makefile(name: str | Path, build_path: Path = None) -> set[Target]:
    ''' Include a sub-directory (or a sub-makefile).
    :returns: The set of exported targets.
    '''
    global context
    if not context.root:
        assert type(name) == type(Path())
        module_path: Path = name / 'dan-build.py'
        spec = importlib.util.spec_from_file_location(
            'root', module_path)
        name = 'root'
    else:
        lookups = [
            os.path.join(name, 'dan-build.py'),
            f'{name}.py',
        ]
        for lookup in lookups:
            module_path = context.current.source_path / lookup
            if module_path.exists():
                spec = importlib.util.spec_from_file_location(
                    f'{context.current.name}.{name}', module_path)
                break
        else:
            raise RuntimeError(
                f'Cannot find anything to include for "{name}" (looked for: {", ".join(lookups)})')
        
    if module_path in context.imported_makefiles:
        return context.imported_makefiles[module_path]

    module = importlib.util.module_from_spec(spec)
    context.imported_makefiles[module_path] = module
    _init_makefile(module, name, build_path)

    requirements_file = module_path.with_stem('dan-requires')
    if module_path.stem == 'dan-build' and requirements_file.exists():
        context.current.requirements = load_makefile(
            requirements_file, name='dan-requires', module_name=f'{name}.requirements')

    try:
        spec.loader.exec_module(module)
    except TargetNotFound as err:
        if len(context.missing) == 0:
            raise err
    except Exception as err:
        context.error('makefile error while including \'%s\': %s', module_path, err)
        raise MakeFileError(module_path) from err
    context.up()


def include(*names: str | Path) -> list[Target]:
    """Include one (or more) subdirectory (or named makefile).

    :param names: One (or more) subdirectory or makefile to include.
    :return: The list of targets exported by the included targets.
    """
    for name in names:
        include_makefile(name)



import os
from dan.core.pathlib import Path
import subprocess
import sys
import tempfile
import functools

import json
import pickle
from dan.core.find import find_executable

from dan import logging
from dan.core.runners import sync_run
from dan.core.version import Version
from dan.core.win import vswhere


class CompilerId:
    def __init__(self, name: str, version: Version, arch: str, system: str) -> None:
        self.name = name
        self.version = version
        self.arch = arch
        self.system = system

    def __str__(self) -> str:
        return f'{self.name}-{self.version}'
    
    @property
    def is_unix(self):
        return self.system not in ('windows', )

GCC = "gcc"
LLVM_GCC = "llvm-gcc"  # GCC frontend with LLVM backend
CLANG = "clang"
APPLE_CLANG = "apple-clang"
SUNCC = "suncc"
VISUAL_STUDIO = "Visual Studio"
INTEL = "intel"
QCC = "qcc"
MCST_LCC = "mcst-lcc"
MSVC = "msvc"

MSVC_TO_VS_VERSION = {800: (1, 0),
                      900: (2, 0),
                      1000: (4, 0),
                      1010: (4, 1),
                      1020: (4, 2),
                      1100: (5, 0),
                      1200: (6, 0),
                      1300: (7, 0),
                      1310: (7, 1),
                      1400: (8, 0),
                      1500: (9, 0),
                      1600: (10, 0),
                      1700: (11, 0),
                      1800: (12, 0),
                      1900: (14, 0),
                      1910: (15, 0),
                      1911: (15, 3),
                      1912: (15, 5),
                      1913: (15, 6),
                      1914: (15, 7),
                      1915: (15, 8),
                      1916: (15, 9),
                      1920: (16, 0),
                      1921: (16, 1),
                      1922: (16, 2),
                      1923: (16, 3),
                      1924: (16, 4),
                      1925: (16, 5),
                      1926: (16, 6),
                      1927: (16, 7),
                      1928: (16, 8),
                      1929: (16, 10),
                      1930: (17, 0)}

_arm_defines = {
    'arm2': ('__ARM_ARCH_2__'),
    'arm3': ('__ARM_ARCH_3__', '__ARM_ARCH_3M__'),
    'arm4': ('__ARM_ARCH_4T__', '__TARGET_ARM_4T'),
    'arm5': ('__ARM_ARCH_5_', '__ARM_ARCH_5E_'),
    'arm6': ('__ARM_ARCH_6T2_', '__ARM_ARCH_6T2_', '__ARM_ARCH_6__', '__ARM_ARCH_6J__', '__ARM_ARCH_6K__', '__ARM_ARCH_6Z__', '__ARM_ARCH_6ZK__'),
    'arm7': ('__ARM_ARCH_7__', '__ARM_ARCH_7A__', '__ARM_ARCH_7R__', '__ARM_ARCH_7M__', '__ARM_ARCH_7S__'),
    'arm7a': ('__ARM_ARCH_7A__', '__ARM_ARCH_7R__', '__ARM_ARCH_7M__', '__ARM_ARCH_7S__'),
    'arm7r': ('__ARM_ARCH_7R__', '__ARM_ARCH_7M__', '__ARM_ARCH_7S__'),
    'arm7m': ('__ARM_ARCH_7M__'),
    'arm7s': ('__ARM_ARCH_7S__'),
}

def dict_contains(defines, *defs):
    for d in defs:
        if d in defines:
            return True
    return False

def get_target_system(defines: dict[str, str]) -> str:
    system = None
    if dict_contains(defines, '__MSYS__', '__MINGW32__'):
        if dict_contains(defines, '__MINGW64__'):
            system = 'msys-mingw64'
        elif dict_contains(defines, '__MINGW32__'):
            system = 'msys-mingw32'
        else:
            system = 'msys'
    elif dict_contains(defines, '_WIN32', '_WIN64'):
        system = 'windows'
    elif dict_contains(defines, '__ANDROID__'):
        system = 'android'
    elif dict_contains(defines, '__linux__'):
        system = 'linux'
    elif dict_contains(defines, '__sun'):
        system = 'sun'
    elif dict_contains(defines, '__hpux'):
        system = 'hpux'
    elif dict_contains(defines, '__DragonFly__'):
        system = 'dragonfly'
    elif dict_contains(defines, '__FreeBSD__'):
        system = 'freebsd'
    elif dict_contains(defines, '__NetBSD__'):
        system = 'netbsd'
    elif dict_contains(defines, '__OpenBSD__'):
        system = 'openbsd'
    elif dict_contains(defines, 'BSD'):
        system = 'bsd'
    elif dict_contains(defines, '__unix__'):
        system = 'unix'
    elif dict_contains(defines, '__MACH__', '__APPLE__'):
        system = 'macos'
    return system

def get_target_arch(defines: dict[str, str]) -> str:
    arch = None
    if dict_contains(defines, '__x86_64__', '_M_X64'):
        arch = 'x64'
    elif dict_contains(defines, 'i386', '__i386__', '__i386', '_M_IX86'):
        arch = 'x86'
    elif dict_contains(defines, '__aarch64__', '_M_ARM64'):
        arch = 'arm64'
    elif '_M_ARM' in defines:
        arch = 'arm'
    else:
        for defs in _arm_defines:
            if dict_contains(defines, *defs):
                arch = 'arm'
                break
    return arch

def _parse_compiler_version(defines: dict[str, str]):
    try:
        if '__LCC__' in defines and '__e2k__' in defines:
            compiler = MCST_LCC
            version = int(defines['__LCC__'])
            major = int(version / 100)
            minor = int(version % 100)
            patch = int(defines['__LCC_MINOR__'])
        elif '__INTEL_COMPILER' in defines:
            compiler = INTEL
            version = int(defines['__INTEL_COMPILER'])
            major = int(version / 100)
            minor = int(version % 100)
            patch = int(defines['__INTEL_COMPILER_UPDATE'])
        elif '__clang__' in defines:
            compiler = APPLE_CLANG if '__apple_build_version__' in defines else CLANG
            major = int(defines['__clang_major__'])
            minor = int(defines['__clang_minor__'])
            patch = int(defines['__clang_patchlevel__'])
        elif '__SUNPRO_C' in defines or '__SUNPRO_CC' in defines:
            # In particular, the value of __SUNPRO_CC, which is a three-digit hex number.
            # The first digit is the major release. The second digit is the minor release.
            # The third digit is the micro release. For example, C++ 5.9 is 0x590.
            compiler = SUNCC
            define = '__SUNPRO_C' if '__SUNPRO_C' in defines else '__SUNPRO_CC'
            version = int(defines[define], 16)
            major = (version >> 8) & 0xF
            minor = (version >> 4) & 0xF
            patch = version & 0xF
        # MSVC goes after Clang and Intel, as they may define _MSC_VER
        elif '_MSC_VER' in defines:
            version = int(defines['_MSC_VER'])
            full_version = 0
            if '_MSC_FULL_VER' in defines:
                full_version = int(defines['_MSC_FULL_VER'])
            # Visual Studio 2022 onwards, detect as a new compiler "msvc"
            if version >= 1930:
                compiler = MSVC
                major = int(version / 100)
                minor = int(version % 100)
                patch = int(full_version % 100000)
            else:
                compiler = VISUAL_STUDIO
                # map _MSC_VER into conan-friendly Visual Studio version
                # currently, conan uses major only, but here we store minor for the future as well
                # https://docs.microsoft.com/en-us/cpp/preprocessor/predefined-macros?view=vs-2019
                major, minor = MSVC_TO_VS_VERSION.get(version)
                # special cases 19.8 and 19.9, 19.10 and 19.11
                if (major, minor) == (16, 8) and full_version >= 192829500:
                    major, minor = 16, 9
                if (major, minor) == (16, 10) and full_version >= 192930100:
                    major, minor = 16, 11
                patch = 0
        # GCC must be the last try, as other compilers may define __GNUC__ for compatibility
        elif '__GNUC__' in defines:
            if '__llvm__' in defines:
                compiler = LLVM_GCC
            elif '__QNX__' in defines:
                compiler = QCC
            else:
                compiler = GCC
            major = int(defines['__GNUC__'])
            minor = int(defines['__GNUC_MINOR__'])
            patch = int(defines['__GNUC_PATCHLEVEL__'])
        else:
            return None
        
        system = get_target_system(defines)
        arch = get_target_arch(defines)

        return CompilerId(compiler, Version(major, minor, patch), arch=arch, system=system)
    except KeyError:
        return None
    except ValueError:
        return None
    except TypeError:
        return None

__data_path = Path(__file__).parent / 'data'
__empty_source = __data_path / 'empty.c'
__detect_cmd = __data_path / 'detect.cmd'

detectors = {
    # "-dM" generate list of #define directives
    # "-E" run only preprocessor
    # "-x c" compiler as C code
    # the output is of lines in form of "#define name value"
    'gcc': ['-dM', '-E', '-x', 'c'],
    'clang': ['-dM', '-E', '-x', 'c'],
    'clang-cl': ['--driver-mode=g++', '-dM', '-E', '-x', 'c'],
    'sun-cc': ['-c', '-xdumpmacros'],
    # cl (Visual Studio, MSVC)
    # "/nologo" Suppress Startup Banner
    # "/E" Preprocess to stdout
    # "/B1" C front-end
    # "/c" Compile Without Linking
    # "/TC" Specify Source File Type
    'msvc': ['/nologo', '/E', '/B1', str(__detect_cmd), '/c', '/TC'],
    'icc': ['/QdM', '/E', '/TC'],  # icc (Intel) on Windows,
    'qcc': ['-Wp', '-dM', '-E', '-x', 'c'],  # QNX QCC
}

def parse_compiler_defines(output: str):
    defines = dict()
    for line in output.splitlines():
        tokens = line.split(' ', 3)
        if len(tokens) == 3 and tokens[0] == '#define':
            defines[tokens[1]] = tokens[2]
        # MSVC dumps macro definitions in single line:
        # "MSC_CMD_FLAGS=-D_MSC_VER=1921 -Ze"
        elif line.startswith("MSC_CMD_FLAGS="):
            line = line[len("MSC_CMD_FLAGS="):].rstrip()
            defines = dict()
            tokens = line.split()
            for token in tokens:
                if token.startswith("-D") or token.startswith("/D"):
                    token = token[2:]
                    if '=' in token:
                        name, value = token.split('=', 2)
                    else:
                        name, value = token, '1'
                    defines[name] = value
            break
    return defines

def get_compiler_defines(executable: str, compiler_type: str, options: list[str], env=None) -> dict[str, str]:
    if env is None:
        env = dict()
    env['LC_LOCAL'] = 'C'
    with tempfile.TemporaryDirectory(prefix='dan-dci-') as tmpdir:
        output, _, rc = sync_run(
                [executable, *detectors[compiler_type], *options, str(__empty_source)], env=env, cwd=tmpdir)
        return parse_compiler_defines(output)


def detect_compiler_id(executable, env=None, logger=None):
    # use a temporary file, as /dev/null might not be available on all platforms
    with tempfile.TemporaryDirectory(prefix='dan-dci-') as tmpdir:
        for name, detector in detectors.items():
            output, _, rc = sync_run(
                [executable, *detector, str(__empty_source)], no_raise=True, env=env, cwd=tmpdir, logger=logger)
            if 0 == rc:
                defines = parse_compiler_defines(output)
                compiler = _parse_compiler_version(defines)
                if compiler is None:
                    continue
                return compiler
        return None


class Compiler:
    def __init__(self, path: Path, env: dict[str, str] = None, tools: dict[str, Path] = dict(), logger=None) -> None:
        self.path = path
        if env is None:
            env = dict()
        env['LC_LOCAL'] = 'C'
        epath = env.get('PATH', os.environ['PATH']).split(os.pathsep)
        if not str(path.parent) in epath:
            epath.insert(0, str(path.parent))
        env['PATH'] = os.pathsep.join(epath)
        self.compiler_id = detect_compiler_id(path, env=env, logger=logger)
        if self.compiler_id is None:
            raise RuntimeError(f'Cannot detect compiler ID of {self.path}')
        self.name = self.compiler_id.name
        self.env = env
        self.tools = tools

    @property
    def arch(self):
        return self.compiler_id.arch

    @property
    def system(self):
        return self.compiler_id.system
    
    @property
    def version(self):
        return self.compiler_id.version
    
    @property
    def defines(self):
        return self.compiler_id.defines

    def __str__(self) -> str:
        return f'{self.compiler_id} {self.arch + " " if self.arch else ""}({self.path})'

    def __eq__(self, other: 'Compiler'):
        return self.path == other.path and self.arch == other.arch and self.system == other.system

    def __hash__(self):
        return hash(self.path) ^ hash(self.arch) ^ hash(self.system)


def validate_pair(ob):
    try:
        if not (len(ob) == 2):
            print("Unexpected result:", ob, file=sys.stderr)
            raise ValueError
    except:
        return False
    return True


def get_environment_from_batch_command(env_cmd, initial=None):
    """
    Take a command (either a single command or list of arguments)
    and return the environment created after running that command.
    Note that if the command must be a batch file or .cmd file, or the
    changes to the environment will not be captured.

    If initial is supplied, it is used as the initial environment passed
    to the child process.
    """
    def consume(iter):
        try:
            while True:
                next(iter)
        except StopIteration:
            pass
    if not isinstance(env_cmd, (list, tuple)):
        env_cmd = [env_cmd]
    # construct the command that will alter the environment
    env_cmd = subprocess.list2cmdline(env_cmd)
    # create a tag so we can tell in the output when the proc is done
    if os.name == 'nt':
        # construct a cmd.exe command to do accomplish this
        cmd = f'cmd.exe /s /c "{env_cmd} && set"'
        enc = 'cp1252'
        shell = False
    else:
        cmd = f"sh -c '. {env_cmd} && env'"
        enc = 'utf-8'
        shell = True
    # launch the process
    proc = subprocess.Popen(cmd, stdout=subprocess.PIPE,
                            env=initial, shell=shell)
    # parse the output sent to stdout
    out, _ = proc.communicate()
    result = dict()
    for line in out.decode(enc).splitlines():
        line = line.strip()
        pos = line.find('=')
        if pos > 0:
            name = line[:pos].strip()
            value = line[pos + 1:].strip()
            # if os.getenv(name) != value:
            #    result[name] = value
            result[name] = value
    return result


def get_compilers(logger: logging.Logger, paths = None):
    if paths is None:
        default_paths = True
    else:
        default_paths = False
    from dan.core.find import find_executables, find_executable, find_file
    compilers: set[Compiler] = set()
    if paths is None and os.name == 'nt':
        infos = vswhere()
        for info in infos:
            logger.info(f'Loading Visual Studio: {info["displayName"]}')
            paths = [info['installationPath']]
            vcvars = find_file(r'vcvarsall.bat$', paths=paths)
            archs = [('x86_64', 'x64'), ('x86', 'x86')]
            for arch, vc_arch in archs:
                logger.debug('Loading Visual Studio environment: %s (%s)', vcvars, vc_arch)
                env = get_environment_from_batch_command([vcvars, vc_arch])
                logger.debug('Visual Studio environment: %s', env)
                paths = env.get('PATH', env.get('Path', None))
                if paths is None:
                    raise RuntimeError('Cannot get PATH in Visual Studio environment')
                paths = paths.split(os.pathsep)
                cl = find_executable('cl', paths=paths, default_paths=False)
                link = find_executable(
                    'link', paths=[cl.parent], default_paths=False)
                lib = find_executable(
                    'lib', paths=[cl.parent], default_paths=False)
                if env:
                    cc = Compiler(cl, env=env,
                                  tools={'link': link, 'lib': lib}, logger=logger)
                    assert cc.arch == vc_arch
                    compilers.add(cc)
                else:
                    logger.warning(
                        f'Cannot load msvc with {arch} architecture')
    else:
        logger.debug('looking for gcc%s', logging.lazy_fmt(lambda: '' if paths is None else ' in' + ', '.join(paths)))
        for gcc in find_executables(r'gcc(-\d+)?(\.exe)?', paths, default_paths):
            gcc = gcc.resolve()
            compilers.add(Compiler(gcc, logger=logger))
        logger.debug('looking for clang%s', logging.lazy_fmt(lambda: '' if paths is None else ' in' + ', '.join(paths)))
        for clang in find_executables(r'clang(-\d+)?(\.exe)?', paths, default_paths):
            clang = clang.resolve()
            compilers.add(Compiler(clang, logger=logger))
    return compilers

unix_tools = [
    'nm', 'ranlib', 'strip', 'readelf', 'ar', 'ranlib', ('dbg', 'gdb')
]

if os.name != 'nt':
    _required_tools = unix_tools
else:
    _required_tools = list()


def create_toolchain(compiler: Compiler, logger=logging.getLogger('toolchain')):
    logger.info('scanning %s toolchain (%s-%s)', compiler.compiler_id, compiler.system, compiler.arch)
    data = {
        'type': compiler.name,
        'version': str(compiler.version),
        'cc': str(compiler.path),
    }
    if len(compiler.path.suffixes):
        extension = compiler.path.suffixes[-1]
    else:
        extension = ''
    pos = compiler.path.stem.rfind(compiler.name)
    if pos >= 0:
        prefix = None if pos == 0 else compiler.path.stem[:pos]
        base_name = compiler.name
        suffix = compiler.path.stem[pos + len(compiler.name):]
    else:
        prefix = None
        base_name = compiler.path.stem
        suffix = None
    base_path = compiler.path.parent

    data['arch'] = compiler.arch
    data['system'] = compiler.system

    if compiler.env:
        data['env'] = compiler.env

    def get_compiler_tool(tool, toolname=None):
        if isinstance(tool, tuple):
            tool, toolname = tool
        if not toolname:
            toolname = f'{base_name}-{tool}'
        if prefix:
            toolname = f'{prefix}{toolname}'
        if suffix:
            toolname = f'{toolname}{suffix}'
        tool_path = (base_path / toolname).with_suffix(extension)
        if tool_path.exists():
            logger.debug(f'found {tool} tool: {tool_path}')
            data[tool] = str(tool_path)
        else:
            logger.debug(f'{tool} tool not found: {tool_path}')
    if compiler.name == 'gcc':
        get_compiler_tool('cxx', 'g++')
        get_compiler_tool('as')
        get_compiler_tool('dbg', 'gdb')
    elif compiler.name == 'clang':
        get_compiler_tool('cxx', 'clang++')
        get_compiler_tool('dbg', 'lldb')
    elif compiler.name == 'msvc':
        data['link'] = str(compiler.tools['link'])
        data['lib'] = str(compiler.tools['lib'])

    if compiler.compiler_id.is_unix:
        for tool in unix_tools:
            get_compiler_tool(tool)

    name = str(compiler.compiler_id)
    if prefix:
        name = f'{prefix}{name}'
    if suffix:
        name = f'{name}{suffix}'
    return name, data


_home_var = 'USERPROFILE' if os.name == 'nt' else 'HOME'


@functools.cache
def get_dan_path():
    path = Path(os.getenv('DAN_DATA', os.getenv(_home_var))) / '.dan'
    path.mkdir(exist_ok=True, parents=False)
    return path


def get_toolchain_path():
    return get_dan_path() / 'toolchains.dat'


def load_env_toolchain(script: Path = None, name: str = None):
    logger = logging.getLogger('toolchain')
    env = get_environment_from_batch_command(script)
    arch = env['ARCH'] if 'ARCH' in env else None
    paths = env['PATH'].split(os.pathsep)

    def patch_flags(name, flagsname):
        parts = env[name].split(' ')
        env[name] = parts[0]
        env[flagsname] = ' '.join(
            set([*env[flagsname].split(' '), *parts[1:]]))
    # split CC/CFLAGS
    patch_flags('CC', 'CFLAGS')
    patch_flags('CXX', 'CXXFLAGS')
    patch_flags('LD', 'LDFLAGS')
    if 'CPP' in env:
        del env['CPP']
    if 'CPPFLAGS' in env:
        del env['CPPFLAGS']

    cc_path = find_executable(env['CC'], paths, default_paths=False)
    cc = Compiler(Path(cc_path), arch=arch, env=env)
    tname, toolchain = create_toolchain(cc, logger)
    save_toolchain(name or tname, toolchain)


def save_toolchain(name, toolchain):
    toolchains_path = get_toolchain_path()
    logger = logging.getLogger('toolchain')
    if toolchains_path.exists():
        with open(toolchains_path, 'rb+') as f:
            logger.info(f'updating toolchains file {toolchains_path}')
            data = pickle.load(f)
            toolchains = data['toolchains']
            toolchains[name] = toolchain
            f.seek(0)
            f.truncate()
            pickle.dump(data, f)


def create_toolchains(paths = None):
    if paths is None:
        default_paths = True
    else:
        default_paths = False
    from dan.core.find import find_executable
    toolchains_path = get_toolchain_path()
    logger = logging.getLogger('toolchain')
    data = None
    if toolchains_path.exists():
        with open(toolchains_path, 'rb') as f:
            logger.info(f'updating toolchains file {toolchains_path}')
            data = pickle.load(f)
            if data:
                toolchains = data['toolchains']
                tools = data['tools']
    if not data:
        data = dict()
        toolchains = dict()
        tools = dict()

    compilers = get_compilers(logger, paths)
    if len(compilers) == 0:
        logger.warning('no toolchain found')
        return data
    for cc in compilers:
        k, v = create_toolchain(cc, logger)
        arch_k = f'{k}-{v["arch"]}'
        if arch_k in toolchains.keys():
            k = arch_k
        if not k in toolchains.keys():
            logger.info(f'new toolchain \'{k}\' found')
        elif toolchains[k] != v:
            if toolchains[k]['type'] == v['type'] and toolchains[k]['arch'] != v['arch']:                
                # add arch suffix
                old_arch = toolchains[k]['arch']
                logger.info(f'renaming \'{k}\' -> \'{k}-{old_arch}\'')
                toolchains[f'{k}-{old_arch}'] = toolchains.pop(k)
                logger.info(f'new toolchain \'{arch_k}\' found')
                toolchains[f'{arch_k}'] = v
                continue
            else:
                logger.info(f'updating toolchain \'{k}\'')
        else:
            logger.info(f'toolchain \'{k}\' unchanged')
            continue
        toolchains[k] = v
    for tool in _required_tools:
        if isinstance(tool, tuple):
            tool, toolname = tool
        else:
            toolname = tool
        tools[tool] = str(find_executable(toolname, paths, default_paths))
    data['tools'] = tools
    data['toolchains'] = toolchains
    if not 'default' in data:
        from dan.core.osinfo import OSInfo
        osi = OSInfo()
        default_toolchain = None
        for name, toolchain in toolchains.items():
            if toolchain['system'] == osi.name and toolchain['arch'] == osi.arch:
                default_toolchain = name
                break
        if default_toolchain is None:
            default_toolchain = list(toolchains.keys())[0]
        logger.debug('selected default toolchain: %s', default_toolchain)
        data['default'] = default_toolchain

    json_toolchain_path = toolchains_path.with_suffix('.json')
    with open(json_toolchain_path, 'w') as jf, open(toolchains_path, 'wb') as pf:
        pickle.dump(data, pf)
        json.dump(data, jf, indent=4)
    return data


def get_toolchains(create = True):
    toolchains_path = get_toolchain_path()
    if not toolchains_path.exists():
        if not create:
            return {'toolchains': dict()}
        return create_toolchains()
    
    json_toolchain_path = toolchains_path.with_suffix('.json')

    with open(toolchains_path, 'rb') as f:
        data = pickle.load(f)
    
    # pickle/json synchronization
    if not json_toolchain_path.exists() or json_toolchain_path.older_than(toolchains_path):
        logger = logging.getLogger('toolchain')
        logger.info(f'Updating {json_toolchain_path.name}')
        with open(json_toolchain_path, 'w') as f:
            json.dump(data, f, indent=4)
            toolchains_path.touch()
    elif json_toolchain_path.exists() and json_toolchain_path.younger_than(toolchains_path):
        logger = logging.getLogger('toolchain')
        logger.info(f'Updating {toolchains_path.name} ({json_toolchain_path.name} changed)')
        with open(json_toolchain_path, 'r') as jf, open(toolchains_path, 'wb') as pf:
            data = json.load(jf)
            pickle.dump(data, pf)

    return data



import os

from dan.core import aiofiles, asyncio
from dan.core.pathlib import Path
from dan.core.settings import InstallMode, InstallSettings
from dan.core.target import Target
from dan.core.find import find_file, find_files
from dan.core.version import Version, VersionSpec
from dan.io.repositories import get_packages_path, get_repo_instance


class PackageBuild(Target, internal=True):

    _all_builds: dict[str, 'PackageBuild'] = dict()
    
    def __init__(self, name, version, package, repository, *args, spec: VersionSpec = None, **kwargs):
        self.spec = spec
        self.pn = name
        super().__init__(f'{name}-build', *args, version=version, **kwargs)
        self.package = name if package is None else package
        self.repo = get_repo_instance(repository, self.makefile)
        self.preload_dependencies.add(self.repo)
        self._package_makefile = None
        self._build_path = None

    @property
    def package_makefile(self):
        if self._package_makefile is None:
            target = self.repo.find(self.pn, self.package)
            if target is None:
                raise RuntimeError(f'cannot find {self.pn} in {self.repo.name}')
            self._package_makefile = target.makefile
        return self._package_makefile

    def get_sources(self):
        makefile = self.package_makefile
        sources = None
        for target in makefile.all_targets:
            if 'source' in target.name:
                sources = target
                break
        if sources is None:
            raise RuntimeError(f'Cannot find {self.pn} package\'s sources target')
        return sources

    async def __initialize__(self):
        sources = self.get_sources()
        if self.spec is not None:
            avail_versions = await sources.available_versions()
            if avail_versions is None:
                self.warning(f'unable to get available versions, default one will be used')
            else:
                for version in avail_versions:
                    if self.spec.is_compatible(version):
                        self.debug(f'using version {version} to match {self.spec}')
                        self.version = version
                        break

        packages_path = get_packages_path()
        makefile = self.package_makefile

        # set package version
        version_option = makefile.options.get('version')
        if self.version is None:
            self.version = Version(version_option.value)
        else:
            version_option.value = str(self.version)

        toolchain = self.context.get('cxx_target_toolchain')
        self._build_path = packages_path / toolchain.system / toolchain.arch / toolchain.build_type.name / self.package / str(self.version)
        self.install_settings = InstallSettings(self.build_path)
        
        # update package build-path
        makefile.build_path = self.build_path / 'build'


        # set our output to the last installed package
        # TODO handle multiple outputs, then set our outputs to all installed packages
        pkg_name = makefile.all_installed[-1].name     
        self.output = Path(self.install_settings.libraries_prefix) / 'pkgconfig' / f'{pkg_name}.pc'
        sources.output = self.build_path / 'src' # TODO source_prefix in install settings

        return await super().__initialize__()
    
    @property
    def build_path(self) -> Path:
        return self._build_path
    
    async def __build__(self):
        ident = f'{self.package}-{self.version}'
        if ident in self._all_builds:
            self.debug(f'{ident} already built by {self._all_builds[ident].fullname}')
            await self._all_builds[ident].build()
            return

        self._all_builds[ident] = self

        makefile = self.package_makefile
        build_path = makefile.build_path

        # FIXME: shall a makefile have an associated toolchain ?
        toolchain = None
        async with asyncio.TaskGroup(f'installing {self.package}\'s targets') as group:
            for target in makefile.all_installed:
                if hasattr(target, 'toolchain'):
                    if toolchain is None:
                        toolchain = target.toolchain
                    else:
                        assert toolchain == target.toolchain, 'Toolchain missmatch'
                group.create_task(target.install(self.install_settings, InstallMode.dev))

        makefile.cache.ignore()
        del makefile

        os.chdir(self.build_path.parent)

        self.debug('cleaning')
        async with asyncio.TaskGroup(f'cleanup {self.package}') as group:
            if toolchain is not None and not toolchain.build_type.is_debug_mode:
                group.create_task(aiofiles.rmtree(self.output / 'src'))
            group.create_task(aiofiles.rmtree(build_path, force=True))


class Package(Target, internal=True):

    __all: dict[str, 'Package'] = dict()

    def __init__(self,
                 name: str = None,
                 version: str = None,
                 package: str = None,
                 repository: str = None, **kwargs) -> None:
        self.package = package
        self.repository = repository
        if version is not None:
            self.version = version
        if name is not None:
            self.name = name
        super().__init__(**kwargs)
        if self.name in self.__all:
            raise RuntimeError(f'duplicate package: {self.name}')
        self.__all[self.name] = self

    @classmethod
    def instance(cls, name, version, *args, **kwargs):
        if name in cls.__all:
            pkg = cls.__all[name]
            if not version.is_compatible(pkg.version):
                raise RuntimeError(f'incompatible package version: {pkg.version} {version}')
            return pkg, False
        else:
            return Package(name, version, *args, **kwargs), True

    
    async def __initialize__(self):

        match self.version:
            case str():
                _name, spec = VersionSpec.parse(self.version)
                if _name is not None:
                    name = _name        
                if spec:
                    self.version = spec.version
                    self.spec = spec
                else:
                    self.spec = VersionSpec(Version(self.version), '=')
            case VersionSpec():
                self.spec = self.version
                self.version = self.spec.version
            case Version():
                self.spec = VersionSpec(self.version, '=')
            case None:
                self.spec = None

        self.pkg_build = PackageBuild(self.name,
                                      self.version,
                                      self.package,
                                      self.repository,
                                      spec=self.spec,
                                      makefile=self.makefile)
        self.dependencies.add(self.pkg_build)
        self.pkgconfig_path = Path('pkgs') / 'lib' / 'pkgconfig'
        self.dan_path = Path('pkgs') / 'lib' / 'dan'
        self.output = self.pkgconfig_path / f'{self.name}.pc'
        return await super().__initialize__()
    
    async def __build__(self):
        self.pkgconfig_path.mkdir(exist_ok=True, parents=True)
        self.dan_path.mkdir(exist_ok=True, parents=True)      

        async with asyncio.TaskGroup(f'importing {self.name} package') as group:
            for pkg in find_files(r'.+\.pc$', [self.pkg_build.install_settings.libraries_destination / 'pkgconfig']):
                self.debug('copying %s to %s', pkg, self.build_path / self.pkgconfig_path)
                group.create_task(aiofiles.copy(pkg, self.build_path / self.pkgconfig_path))

            for pkg in find_files(r'.+\.py$', [self.pkg_build.install_settings.libraries_destination / 'dan']):
                self.debug('copying %s to %s', pkg, self.build_path / self.dan_path)
                group.create_task(aiofiles.copy(pkg, self.build_path / self.dan_path))
        
        if self.output.exists():
            from dan.pkgconfig.package import Data, find_package
            data = Data(self.output)
            async with asyncio.TaskGroup(f'importing {self.name} package requirements') as group:
                toolchain = self.context.get('cxx_target_toolchain')
                search_path = get_packages_path() / toolchain.system / toolchain.arch / toolchain.build_type.name
                dest = self.build_path / self.pkgconfig_path
                for pkg in data.requires:
                    pkgconfig_file = find_file(rf'{pkg.name}.pc$', [search_path])
                    # NOTE: find_package will resolve to the build-directory installed pkgconfig, wich will result in a failure
                    # pkg = find_package(pkg.name, spec=pkg.version_spec, search_paths=[search_path], makefile=self.makefile)
                    if pkgconfig_file is not None and not (dest / pkgconfig_file.name).exists():
                        self.debug('copying %s to %s', pkgconfig_file, dest)
                        group.create_task(aiofiles.copy(pkgconfig_file, dest))



import dataclasses
import functools
import json
import pickle
import aiofiles
import typing as t


from dan.core.pathlib import Path
from dan.core import asyncio

T = t.TypeVar('T', bound=dict)

class Cache(t.Generic[T]):
    dataclass: T = dict
    indent = None
    __caches: dict[str, 'Cache'] = dict()

    def __init_subclass__(cls) -> None:
        cls.dataclass = t.get_args(cls.__orig_bases__[0])[0]
        return super().__init_subclass__()

    def __init__(self, path: Path|str, *args, cache_name:str = None, binary=False, **kwargs):
        self.__path = Path(path)     
        self.__name = cache_name or path.stem   
        self.__serializer = json if not binary else pickle
        if self.name in self.__caches:
            other = Cache.get(self.name)
            if other.path == self.path:
                raise RuntimeError(f'Cache {self.name} already created, use Cache.instance')
            else:
                raise RuntimeError(f'Cache {self.name} is not unique, use cache_name to distinguish {other.path} from {self.path}')
        assert not self.name in self.__caches, 'a cache should be unique'
        if self.path.exists():
            with open(self.path, 'rb') as f:
                if dataclasses.is_dataclass(self.dataclass):
                    self.__data = self.dataclass.from_json(f.read())
                else:
                    self.__data = self.__serializer.load(f)
                if not isinstance(self.__data, self.dataclass):
                    self.__data = self.dataclass(**self.__data)
                self.__modification_date = self.path.modification_time
        else:
            self.__data = self.dataclass(*args, **kwargs)
            self.__modification_date = 0.0
        
        self.__initial_state = self._dump()
        self.__dirty = False
        self.__caches[self.name] = self
    
    @classmethod
    def instance(cls, path: Path|str, *args, cache_name:str = None, **kwargs):
        cache_name = cache_name or path.stem
        if cache_name in cls.__caches:
            return cls.__caches[cache_name]
        return cls(path, *args, cache_name=cache_name, **kwargs)

    @classmethod
    def clear_all(cls):
        del cls.__caches
        cls.__caches = dict()
    
    def _dump(self):        
        if self.__serializer is pickle:
            return self.__serializer.dumps(self.data)
        else:
            if self.dataclass == dict:
                return self.__serializer.dumps(self.data).encode()
            else:
                return self.data.to_json(indent=self.indent).encode()

    
    @property
    def path(self):
        return self.__path
    
    @property
    def name(self):
        return self.__name
    
    @property
    def data(self) -> T|dict:
        return self.__data
    
    @property
    def dirty(self):
        if not self.__dirty:
            self.__state = self._dump()
            self.__dirty = self.__initial_state != self.__state
        return self.__dirty
    
    async def save(self, force=False):
        if self.path and (self.dirty or force):
            if self.__state:
                self.path.parent.mkdir(exist_ok=True, parents=True)
                async with aiofiles.open(self.path, 'wb') as f:
                    await f.write(self.__state)
                self.__dirty = False

    @classmethod
    async def save_all(cls):
        async with asyncio.TaskGroup('saving caches') as group:
            for c in cls.__caches.values():
                if c.dirty:
                    group.create_task(c.save())

    @classmethod
    def get(cls, name) -> 'Cache':
        if name in cls.__caches:
            return cls.__caches[name]

    def ignore(self):
        if self.name in self.__caches:
            del self.__caches[self.name]


def once_method(fn):    
    result_name = f'_{fn.__name__}_result'

    @functools.wraps(fn)
    def wrapper(self, *args, **kwds):
        if hasattr(self, result_name):
            return getattr(self, result_name)
        
        result = fn(self, *args, **kwds)
        setattr(self, result_name, result)
        return result

    return wrapper





import sys
import dan.core.include
from dan.core.include import include, requires
from dan.core.generator import generator
from dan.pkgconfig.package import find_package


self : dan.core.include.MakeFile

class __LazySelf(sys.__class__):
    @property
    def self(__):
        return dan.core.include.context.current

sys.modules[__name__].__class__ = __LazySelf




from dan.core import aiofiles, asyncio
from dan.core.find import find_executable
from dan.core.pathlib import Path
from dan.core.runners import async_run
from dan.cxx.targets import CXXObject, Executable, Library, CXXObjectsTarget
from dan.pkgconfig.package import Package


class _QtMoccer:
    
    async def __initialize__(self):
        qt_core = Package(f'Qt{self.qt_major}Core', makefile=self.makefile)
        await qt_core.initialize()
        self.moc = self.makefile.cache.get('moc_executable')
        if not self.moc:
            search_paths = [qt_core.host_bins if hasattr(qt_core, 'host_bins') else qt_core.bindir, qt_core.prefix]
            self.moc = find_executable(
                'moc', paths=search_paths, default_paths=False)
            self.makefile.cache.moc_executable = str(self.moc)
        self.dependencies.add(qt_core)

        self.includes.private.append(self.build_path)

        for module in self.qt_modules:
            pkg = Package(f'Qt{self.qt_major}{module}', makefile=self.makefile)
            await pkg.initialize()
            self.dependencies.add(pkg)

        self.mocs = self.cache.get('mocs', list())
        for moc_name in self.mocs:
            moc_path = self.build_path / moc_name
            self.objs.append(
                CXXObject(moc_path, self))
            self.other_generated_files.add(moc_path)

        await super().__initialize__()

    async def __clean__(self):
        await super().__clean__()
        self.cache['mocs'] = list()

    async def __build__(self):
        moc_objs = list()

        async def do_moc(file: Path):
            moc_name = file.with_suffix('.moc.cpp').name
            moc_file_path = self.build_path / moc_name
            if not moc_file_path.exists() or file.younger_than(moc_file_path):
                if moc_file_path.exists():
                    self.info(f'updating {moc_file_path}')
                else:
                    self.info(f'generating {moc_file_path}')
                out, err, rc = await async_run([self.moc, *self.includes.private, *self.compile_definitions.private, file], logger=self, log=False)
                if rc == 0 and len(out):
                    async with aiofiles.open(moc_file_path, 'w') as f:
                        await f.write(out)
                        if not moc_name in self.mocs:
                            moc_objs.append(
                                CXXObject(moc_file_path, self))
                            self.mocs.append(moc_name)

        # we have to generate objects first in order to get files dependencies,
        # which are used later to moc those dependencies (whenever they are within the same source directory)

        # build existing objects
        await CXXObjectsTarget.__build__(self)

        # generate moc source files
        async with asyncio.TaskGroup(f'moccing {self.name}') as g:
            sources = set()
            for obj in self.objs:
                for dep in [Path(dep) for dep in obj.cache['deps']]:
                    if self.source_path in dep.parents:
                        sources.add(dep)
            for source in sources:
                g.create_task(do_moc(source))
        
        # add moc sources to objects dependencies
        self.objs.extend(moc_objs)

        # update cache
        self.cache['mocs'] = self.mocs

        # continue parent build process
        await super().__build__()


def moc(modules: list[str] = None, major=6):
    if modules is None:
        modules = ['Widgets']
    def decorator(cls):
        from dan.core.include import context
        @context.current.wraps(cls)
        class QtWapped(_QtMoccer, cls):
            qt_modules = modules
            qt_major = major
        return cls
    return decorator



from copy import deepcopy
import jinja2
from dan.core import aiofiles, asyncio
from dan.core.pathlib import Path
import re
import importlib.util

from dan.core.find import find_file, find_files, library_paths_lookup
from dan.core.pm import re_match
from dan.core.requirements import RequiredPackage, parse_requirement
from dan.core.runners import cmdline2list
from dan.core.settings import InstallMode, InstallSettings
from dan.core.utils import unique
from dan.core.version import Version, VersionSpec
from dan.cxx.targets import CXXTarget, Library
from dan.cxx.toolchain import LibraryList

import typing as t


class MissingPackage(RuntimeError):
    def __init__(self, name) -> None:
        super().__init__(f'package {name} not found')


def find_pkg_config(name, paths=list()) -> Path:
    return find_file(fr'.*{name}\.pc$', [*paths, '$PKG_CONFIG_PATH', *library_paths_lookup])


def find_pkg_configs(name, paths=list()) -> t.Generator[Path, None, None]:
    yield from find_files(fr'.*{name}\.pc$', [*paths, '$PKG_CONFIG_PATH', *library_paths_lookup])


def has_package(name,  paths=list()):
    return find_pkg_config(name,  paths) is not None



class Data:
    def __init__(self, path) -> None:
        self._path = Path(path)
        self._items = dict()
        with open(self._path) as f:
            lines = [l for l in [l.strip().removesuffix('\n')
                                 for l in f.readlines()] if len(l)]
            for line in lines:
                m = Data.__split_expr.match(line)
                if m:
                    k = m.group(1).lower()
                    v = m.group(2)
                    self._items[k] = v
        self.__requires = None
        self.__version = None

    __split_expr = re.compile(r'(.+?)[:=](.+)')

    def get(self, name: str, default=None):
        if not name in self._items:
            return default
        value = self._items[name]
        if type(value) == str:
            while True:
                m = re.search(r'\${(\w+)}', value)
                if m:
                    var = m.group(1)
                    value = value.replace(
                        f'${{{var}}}', self.get(var))
                else:
                    break
        return value
    
    @property
    def path(self) -> Path:
        return self._path
    
    @property
    def requires(self) -> list[RequiredPackage]:
        if self.__requires is None:
            self.__requires = list()
            reqs = self.get('requires')
            if reqs is not None:
                reqs = reqs.strip()
                # conan generates invalid requires clause
                # it should be 'comma separated values but it is not'
                #  eg.: 'boost-headers boost-_cmake'
                if any([c in reqs for c in ',=']):
                    # the right way
                    for req in reqs.split(','):
                        req = parse_requirement(req)
                        self.__requires.append(req)
                else:
                    # conan's way
                    for req in reqs.split(' '):
                        req = parse_requirement(req)
                        self.__requires.append(req)
        return self.__requires
    
    @property
    def version(self) -> Version:
        if self.__version is None:
            v = self.get('version')
            if v:
                self.__version = Version(v)
        return self.__version


class Package(CXXTarget, internal=True):
    all: dict[str, 'Package'] = dict()

    default = False

    def __init__(self, name, search_paths: list[str] = list(), config_path: Path = None, data: Data = None, **kwargs) -> None:
        if data is not None:
            self.config_path = data.path
            self.data = data
        else:
            self.config_path = config_path or find_pkg_config(name, search_paths)
            self.data: Data = None
        self.search_paths = search_paths
        if not self.config_path:
            raise MissingPackage(name)
        self.search_paths.insert(0, self.config_path.parent)
        self.pn = name
        self.all[name] = self

        super().__init__(f'{name}-pkgconfig', **kwargs)

        dan_plugin = self.config_path.parent.parent / \
            'dan' / f'{name}.py'
        if dan_plugin.exists():
            spec = importlib.util.spec_from_file_location(
                f'{name}_plugin', dan_plugin)
            module = importlib.util.module_from_spec(spec)
            setattr(module, 'self', self)
            spec.loader.exec_module(module)
        self.__cflags = None
        self.__libs = None
        self.__lib_paths = None
        if not data:
            self.data = Data(self.config_path)


        self.version = self.data.version

    @property
    def modification_time(self):
        return 0.0

    @property
    def found(self):
        return True

    async def __initialize__(self):
        deps = set()
        reqs = self.data.requires
        if reqs:
            async with asyncio.TaskGroup(f'resolving {self.name}\'s requirements') as group:
                for req in reqs:
                    if req.name in self.all and req.is_compatible(self.all[req.name]):
                        dep = self.all[req.name]
                    else:
                        dep = find_package(req.name, req.version_spec, search_paths=self.search_paths, makefile=self.makefile)
                    if dep is None:
                        raise RuntimeError(f'Unresolved requirement: {req}')
                    group.create_task(dep.initialize())
                    deps.add(dep)
        self.includes.public.append(self.data.get('includedir'))
        self.dependencies.update(deps)

    @property
    def cxx_flags(self):
        if self.__cflags is None:
            cflags = self.data.get('cflags')
            if cflags is not None:
                cflags = cmdline2list(cflags)
                cflags = self.toolchain.from_unix_flags(cflags)
            else:
                cflags = list()
            for dep in self.cxx_dependencies:
                cflags.extend(dep.cxx_flags)
            self.__cflags = unique(cflags)
        return self.__cflags

    @property
    def package_dependencies(self):
        return [pkg for pkg in self.dependencies if isinstance(pkg, Package)]

    def __init_libs(self):
        self.__libs = LibraryList()
        self.__lib_paths = set()
        libs = self.data.get('libs')
        if libs is not None:
            libs = cmdline2list(libs)
            libs = self.toolchain.from_unix_flags(libs)
            for l in libs:
                match re_match(l):
                    case r'-l(.+)' as m:
                        self.__libs.add(m[0])
                    case r'-L(.+)' as m:
                        self.__lib_paths.add(m[0])
                    case r'/LIBPATH:(.+)' as m:
                        self.__lib_paths.add(m[0])
                    case r'-Wl,-rpath,(.+)' as m:
                        self.__lib_paths.add(m[0])
                    case _:
                        self.__libs.add(l)
        for pkg in self.package_dependencies:
            self.__lib_paths.update(pkg.lib_paths)
            self.__libs.extend(pkg.libs)
        self.__lib_paths = list(sorted(self.__lib_paths))

    @property
    def lib_paths(self) -> list[str]:
        if self.__lib_paths is None:
            self.__init_libs()
        return self.__lib_paths
    
    @property
    def libs(self):
        if self.__libs is None:
            self.__init_libs()
        return self.__libs

    @asyncio.cached
    async def install(self, settings: InstallSettings, mode: InstallMode) -> list[Path]:
        settings = deepcopy(settings)
        settings.create_pkg_config = False
        return await super().install(settings, mode)

    @property
    def up_to_date(self):
        return True

    @property
    def modification_time(self):
        return self.config_path.modification_time

    def __getattr__(self, name):
        value = self.data.get(name) if self.data is not None else None
        if value is None:
            raise AttributeError(name)
        return value


_jinja_env: jinja2.Environment = None


def find_package(name, spec: VersionSpec = None, search_paths: list = None, makefile = None):
    
    if name in Package.all:
        pkg = Package.all[name]
        if spec and not spec.is_compatible(pkg.version):
            raise RuntimeError(f'incompatible package {name} ({pkg.version} {spec})')
        return pkg
    if makefile is None:
        from dan.core.include import context
        makefile = context.current
    search_paths = search_paths or makefile.pkgs_path
    for config in find_pkg_configs(name, search_paths):
        if spec is not None:
            data = Data(config)
            if spec.is_compatible(data.version):
                return Package(name, data=data, makefile=makefile)
        else:
            return Package(name, config_path=config, makefile=makefile)


def _get_jinja_env():
    global _jinja_env
    if _jinja_env is None:
        _jinja_env = jinja2.Environment(
            loader=jinja2.PackageLoader('dan.pkgconfig'))
    return _jinja_env


async def create_pkg_config(lib: Library, settings: InstallSettings) -> Path:
    dest = settings.libraries_destination / 'pkgconfig' / f'{lib.name}.pc'
    lib.info(f'creating pkgconfig: {dest}')

    requires = list()
    for dep in lib.dependencies:
        match dep:
            case RequiredPackage():
                if dep.version_spec:
                    requires.append(f'{dep.name} {dep.version_spec.op} {dep.version_spec.version}')
                else:
                    requires.append(dep.name)
            case Library():
                requires.append(dep.name)


    libs = list()
    if not lib.interface:
        libs.extend(lib.toolchain.make_libpath_options(
            [Path(f'${{libdir}}/{lib.name}')]))
        libs.extend(lib.toolchain.make_link_options(
            [Path(f'${{libdir}}/{lib.name}')]))
    libs.extend(lib.link_libraries.public)
    libs.extend(lib.link_options.public)
    libs = lib.toolchain.to_unix_flags(libs)

    cflags = lib.compile_definitions.public
    cflags.extend(lib.toolchain.make_include_options(['${includedir}']))
    cflags = lib.toolchain.to_unix_flags(cflags)

    data = _get_jinja_env()\
        .get_template('pkg.pc.jinja2')\
        .render({
            'lib': lib,
            'libs': libs,
            'cflags': cflags,
            'settings': settings,
            'prefix': Path(settings.destination).absolute(),
            'requires': requires
        })
    dest.parent.mkdir(parents=True, exist_ok=True)
    async with aiofiles.open(dest, 'w') as f:
        await f.write(data)
    return dest



from enum import Enum
import dan.core.diagnostics as diag
from dan.core.pathlib import Path
from dan.core.settings import BuildType, ToolchainSettings
from dan.core.target import FileDependency
from dan.core.runners import async_run, sync_run, CommandError
from dan.core.version import Version
from dan.logging import Logging
from dan.cxx.compile_commands import CompileCommands

import typing as t

import tempfile

CommandArgs = list[str|Path]
CommandArgsList = list[CommandArgs]

class RuntimeType(Enum):
    static = 0
    dynamic = 1

class LibraryList:
    def __init__(self, *items: t.Iterable):
        self._lst = list()
        self.extend(items)
    
    def add(self, item):
        if not item in self._lst:
            self._lst.insert(0, item)

    def extend(self, items):
        for item in items:
            try:
                self._lst.remove(item)
            except ValueError:
                pass
        self._lst.extend(items)

    def __iter__(self):
        return iter(self._lst)
    
    def __reversed__(self):
        return reversed(self._lst)


class BaseFailure(RuntimeError):
    def __init__(self, msg: str, err: CommandError, options: set[str], command: str, toolchain: 'Toolchain', diags: list[diag.Diagnostic], target = None) -> None:
        super().__init__(msg)
        self.options = options
        self.command = command
        self.toolchain = toolchain
        self.stdout = err.stdout
        self.stderr = err.stderr
        self.diags = diags
        self.target = target
    

class CompilationFailure(BaseFailure):
    def __init__(self, err: CommandError, sourcefile: Path, options: set[str], command: str, toolchain: 'Toolchain', diags: list[diag.Diagnostic] = [], target = None) -> None:
        super().__init__(f'failed to compile {sourcefile}: {err.stdout}{err.stderr}', err, options, command, toolchain, diags, target)
        self.sourcefile = sourcefile


class LinkageFailure(BaseFailure):
    def __init__(self, err: CommandError, objects: set[Path], options: set[str], command: str, toolchain: 'Toolchain', diags: list[diag.Diagnostic] = [], target = None) -> None:
        super().__init__(f'failed to link {", ".join([str(o) for o in objects])}: {err.stdout}{err.stderr}', err, options, command, toolchain, diags, target)
        self.objects = objects


class Toolchain(Logging):
    def __init__(self, data: dict[str,str], tools: dict, settings: ToolchainSettings, cache: dict = None) -> None:
        self.cc : Path = None
        self.cxx : Path = None
        self.tools = tools
        self._compile_commands: CompileCommands = None
        self.cxx_flags = set()
        self.cpp_std = 17
        self.type = data['type']
        # self.arch = data['arch']
        self.system = data['system']
        self.version = Version(data['version'])
        self.settings = settings
        self.cache = dict() if cache is None else cache
        Logging.__init__(self, f'{self.type}-{self.version}')
        self.env = None
        self.rpath = None
        self._build_type = BuildType.debug
        self.compile_options: list[str] = list()
        self.link_options: list[str] = list()
        self.rpath = None
        self.runtime = RuntimeType.dynamic
        self.build_type = BuildType.debug

    @property
    def arch(self):
        self.__update_cache()
        return self.cache['arch']
    
    @property
    def is_host(self):
        self.__update_cache()
        return self.cache['is_host']
    
    @property
    def up_to_date(self):
        if not 'arch' in self.cache or self.cache['arch'] is None:
            return False
        if not 'arch_detect_flags' in self.cache or self.cache['arch_detect_flags'] != self.settings.cxx_flags:
            return False
        return True

    
    def __update_cache(self):
        if self.up_to_date:
            return

        from dan.cxx.detect import get_compiler_defines, get_target_arch
        defines = get_compiler_defines(self.cc, self.type, self.settings.cxx_flags, self.env)
        arch = get_target_arch(defines)
        self.cache['defines'] = defines
        self.cache['arch'] = arch
        self.cache['arch_detect_flags'] = self.settings.cxx_flags
        
        from dan.core.osinfo import OSInfo
        osi = OSInfo()
        self.cache['is_host'] = self.system == osi.name and self.arch == osi.arch

    async def get_default_defines(self) -> dict[str, str]:
        return self.cache['defines']

    # @property
    # def compile_commands(self):
    #     if not self._compile_commands:
    #         self._compile_commands = CompileCommands()
    #     return self._compile_commands

    def init(self):
        self.__update_cache()

    def has_cxx_compile_options(*opts) -> bool:
        raise NotImplementedError()

    def make_compile_definitions(self, definitions: set[str]) -> list[str]:
        raise NotImplementedError()

    def make_include_options(self, include_paths: set[Path]) -> list[str]:
        raise NotImplementedError()

    def make_libpath_options(self, libraries: set[Path | str]) -> list[str]:
        raise NotImplementedError()

    def make_link_options(self, libraries: set[Path]) -> list[str]:
        raise NotImplementedError()

    def make_library_name(self, basename: str, shared: bool) -> str:
        raise NotImplementedError()

    def make_executable_name(self, basename: str) -> str:
        raise NotImplementedError()

    async def _handle_compile_output(self, lines) -> t.Iterable[diag.Diagnostic]:
        raise NotImplementedError()
    
    async def _handle_link_output(self, lines) -> t.Iterable[diag.Diagnostic]:
        raise NotImplementedError()

    async def scan_dependencies(self, file: Path, options: set[str], build_path: Path) -> set[FileDependency]:
        raise NotImplementedError()

    def compile_generated_files(self, output: Path) -> set[Path]:
        return set()
    
    def debug_files(self, output: Path) -> set[Path]:
        return set()

    def make_compile_commands(self, sourcefile: Path, output: Path, options: set[str]) -> CommandArgsList:
        raise NotImplementedError()
    
    def from_unix_flags(self, flags: list[str]) -> list[str]:
        """Convert flags from unix-style to target-compiler-style"""
        return flags
    
    def to_unix_flags(self, flags: list[str]) -> list[str]:
        """Convert flags from target-compiler-style to unix-style"""
        return flags

    async def compile(self, sourcefile: Path, output: Path, options: set[str], **kwds):
        commands = self.make_compile_commands(sourcefile, output, options)
        diags = []
        if diag.enabled:
            async def capture(stream):
                with stream as lines:
                    async for diag in self._handle_compile_output(lines):
                        diags.append(diag)
            kwds['all_capture'] = capture
        for index, command in enumerate(commands):
            try:
                await self.run(f'compile{index}', output, command, **kwds, cwd=output.parent)
            except CommandError as err:
                raise CompilationFailure(err, sourcefile, options, command, self, diags) from None
        return commands, diags

    def make_link_commands(self, objects: set[Path], output: Path, options: set[str]) -> CommandArgsList:
        raise NotImplementedError()

    async def link(self, objects: set[Path], output: Path, options: set[str], **kwds):
        commands = self.make_link_commands(objects, output, options)
        diags = []
        if diag.enabled:
            async def capture(stream):
                with stream as lines:
                    async for diag in self._handle_link_output(lines):
                        diags.append(diag)
            kwds['all_capture'] = capture
        for index, command in enumerate(commands):
            try:
                await self.run(f'link{index}', output, command, **kwds, cwd=output.parent)
            except CommandError as err:
                raise LinkageFailure(err, objects, options, command, self, diags) from None
        return commands, diags

    def make_static_lib_commands(self, objects: set[Path], output: Path, options: set[str]) -> CommandArgsList:
        raise NotImplementedError()

    async def static_lib(self, objects: set[Path], output: Path, options: set[str], **kwds):
        commands = self.make_static_lib_commands(objects, output, options)
        for index, command in enumerate(commands):
            try:
                await self.run(f'static_lib{index}', output, command, **kwds, cwd=output.parent)
            except CommandError as err:
                raise LinkageFailure(err, objects, options, command, self) from None
        return commands

    def make_shared_lib_commands(self, objects: set[Path], output: Path, options: set[str]) -> tuple[Path, CommandArgsList]:
        raise NotImplementedError()

    async def shared_lib(self, objects: set[Path], output: Path, options: set[str], **kwds):
        commands = self.make_shared_lib_commands(objects, output, options)
        for index, command in enumerate(commands):
            await self.run(f'shared_lib{index}', output, command, **kwds, cwd=output.parent)
        return commands

    async def run(self, name: str, output: Path, args, quiet=False, **kwds) -> tuple[str, str, int]:
        return await async_run(args, env={**(self.env or dict()), 'LC_ALL': 'C'}, logger=self if not quiet else None, **kwds)

    @property
    def cxxmodules_flags(self) -> list[str]:
        ...

    def can_compile(self, source: str, options: set[str] = set(), extension='.cpp'):
        with tempfile.NamedTemporaryFile('w', suffix=extension) as f:
            f.write(source)
            f.flush()
            try:
                fname = Path(f.name)
                sync_run(self.make_compile_commands(fname, fname.with_suffix('.o'), options)[0])
                return True
            except CommandError:
                return False

    def has_include(self, *includes, options: set[str] = set(), extension='.cpp'):
        source = '\n'.join([f'#include {inc}' for inc in includes])
        return self.can_compile(source, options, extension)

    def has_definition(self, *definitions, options: set[str] = set(), extension='.cpp'):
        source = '\n'.join([f'''#ifndef {d}
        #error "{d} is not defined"
        #endif''' for d in definitions])
        return self.can_compile(source, options, extension)
    
    async def get_default_include_paths(self, lang = 'c++') -> list[str]:
        return []



class BaseDecorator:
    """ Base class helper for decorator classes that might be used on instance methods.
    """
    is_method: bool = False

    class _GetBoundedInstanceWrapper:
        """ Callable that store wrapped class instance and decorator instance. """

        def __init__(self, decorator_instance, wrapped_instance):
            self.decorator_instance = decorator_instance
            self.wrapped_instance = wrapped_instance

        def __call__(self, *args, **kwargs):
            return self.decorator_instance(self.wrapped_instance, *args, **kwargs)

        def __getattr__(self, name):
            return getattr(self.decorator_instance, name)

    def __get__(self, wrapped_instance, _owner):
        """ For object method decoration.
            It will detect __get__ as descriptor protocol.
            So __get__ should return the actual method.
        """
        # pass decorator instance and decorated object instance
        self.is_method = True
        return self._GetBoundedInstanceWrapper(self, wrapped_instance)



from dan.core.pathlib import Path
import re
from typing import Any
from dan.core.target import Target

import aiofiles


class ConfigureFile(Target, internal=True):

    input: Path = None
    variables: dict[str, Any] = dict()

    _cmake_define_expr = re.compile(r'#\s?cmakedefine\s+(\w+)\s?(@\w+@)?')
    _define_expr = re.compile(r'#\s?define\s+(\w+)\s+"?(@(\w+)@)"?')

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.__variables = self.variables

    def __setitem__(self, key, value):
        self.__variables[key] = value

    async def __build__(self):
        self.output.parent.mkdir(exist_ok=True, parents=True)
        def gen_define(name, replacement):
            if name in self.__variables:
                if replacement is None:
                    return f'#define {name}\n' if self.__variables[
                        name] else f'/* #undef {name} */\n'
                else:
                    replacement = replacement[1:-1]
                    assert replacement in self.__variables, f'{replacement} is not defined'
                    replacement = self.__variables[replacement]
                    if isinstance(replacement, str):
                        replacement = f'"{replacement}"'
                    return f'#define {name} {replacement}\n'
            else:
                 return f'/* #undef {name} */\n'
        
        async with aiofiles.open(self.input, 'r') as inf, aiofiles.open(self.output, 'w') as outf:
            for line in await inf.readlines():                
                m = self._cmake_define_expr.search(line)
                if m:
                    line = gen_define(m.group(1), m.group(2))
                m = self._define_expr.search(line)
                if m:
                    line = gen_define(m.group(1), m.group(2))

                await outf.write(line)



import json
import os
from shutil import which
import subprocess
from dan.core.runners import sync_run


def _system_registry_key(key, subkey, query):
    from winreg import winreg  # @UnresolvedImport
    try:
        hkey = winreg.OpenKey(key, subkey)
    except (OSError, WindowsError):  # Raised by OpenKey/Ex if the function fails (py3, py2)
        return None
    else:
        try:
            value, _ = winreg.QueryValueEx(hkey, query)
            return value
        except EnvironmentError:
            return None
        finally:
            winreg.CloseKey(hkey)


def is_win64():
    import winreg  # @UnresolvedImport
    return _system_registry_key(winreg.HKEY_LOCAL_MACHINE,
                                r"SOFTWARE\Microsoft\Windows\CurrentVersion",
                                "ProgramFilesDir (x86)") is not None


def vswhere(all_=False, prerelease=False, products=None, requires=None, version="", latest=False,
            legacy=False, property_="", nologo=True):

    # 'version' option only works if Visual Studio 2017 is installed:
    # https://github.com/Microsoft/vswhere/issues/91

    products = list() if products is None else products
    requires = list() if requires is None else requires

    if legacy and (products or requires):
        raise RuntimeError("The 'legacy' parameter cannot be specified with either the "
                           "'products' or 'requires' parameter")

    installer_path = None
    program_files = os.getenv("ProgramFiles(x86)") or os.getenv("ProgramFiles")
    if program_files:
        expected_path = os.path.join(program_files, "Microsoft Visual Studio", "Installer",
                                     "vswhere.exe")
        if os.path.isfile(expected_path):
            installer_path = expected_path
    vswhere_path = installer_path or which("vswhere")

    if not vswhere_path:
        raise RuntimeError("Cannot locate vswhere in 'Program Files'/'Program Files (x86)' "
                           "directory nor in PATH")

    arguments = list()
    arguments.append(vswhere_path)

    # Output json format
    arguments.append("-format")
    arguments.append("json")

    if all_:
        arguments.append("-all")

    if prerelease:
        arguments.append("-prerelease")

    if products:
        arguments.append("-products")
        arguments.extend(products)

    if requires:
        arguments.append("-requires")
        arguments.extend(requires)

    if len(version) != 0:
        arguments.append("-version")
        arguments.append(version)

    if latest:
        arguments.append("-latest")

    if legacy:
        arguments.append("-legacy")

    if len(property_) != 0:
        arguments.append("-property")
        arguments.append(property_)

    if nologo:
        arguments.append("-nologo")

    try:
        output, err, rc = sync_run(arguments, no_raise=True)
        if rc != 0:
            raise RuntimeError(err.strip())
        output = output.strip()
        # Ignore the "description" field, that even decoded contains non valid charsets for json
        # (ignored ones)
        output = "\n".join([line for line in output.splitlines()
                            if not line.strip().startswith('"description"')])

    except (ValueError, subprocess.CalledProcessError, UnicodeDecodeError) as e:
        raise RuntimeError("vswhere error: %s" % str(e))

    return json.loads(output)



import os
import re
import typing as t

from collections.abc import Iterable
from enum import Enum
from functools import cached_property

from dan.core.pathlib import Path
from dan.core import aiofiles, cache
from dan.core.settings import InstallMode, InstallSettings
from dan.core.target import Target
from dan.core.utils import chunks, unique
from dan.core.runners import async_run
from dan.core import asyncio
from dan.cxx.toolchain import CompilationFailure, LibraryList, LinkageFailure, Toolchain


class CXXObject(Target, internal=True):
    def __init__(self, source:Path, parent: 'CXXTarget', root: Path = None) -> None:
        if source.is_absolute():
            if root is None:
                root = parent.build_path
            name = '-'.join(source.relative_to(root).with_suffix(f'').parts)
        else:
            name = '-'.join(source.with_suffix(f'').parts)
        super().__init__(name, parent=parent, default=False)
        self.parent = parent
        self.source = source
        self.toolchain: Toolchain = self.context.get('cxx_target_toolchain')
        obj_fname = source.with_suffix('.obj' if self.toolchain.type == 'msvc' else '.o')
        if source.is_absolute():
            self.output = self.build_path / obj_fname.name
        else:
            self.output = self.build_path / obj_fname
        self.__dirty = False

    @property
    def cxx_flags(self):
        return self.parent.cxx_flags

    @property
    def private_cxx_flags(self):
        return self.parent.private_cxx_flags
    
    @property
    def includes(self):
        return self.parent.includes

    @property
    def compile_definitions(self):
        return self.parent.compile_definitions

    async def __initialize__(self):
        await self.parent.preload()

        deps = self.cache.get('deps')
        if deps is not None:
            self.dependencies.update(deps)

        self.dependencies.add(self.source)

        self.other_generated_files.update(
            self.toolchain.compile_generated_files(self.output))

        previous_args = self.cache.get('compile_args')
        if previous_args:
            args = self.toolchain.make_compile_commands(
                self.source_path / self.source, self.output, self.private_cxx_flags)[0]
            args = [str(arg) for arg in args]
            if sorted(args) != sorted(previous_args):
                self.__dirty = True
        else:
            self.__dirty = True

    @cached_property
    def up_to_date(self):
        if self.__dirty:
            return False
        return super().up_to_date

    async def __build__(self):
        self.info('generating %s...', self.output.name)
        try:
            self.output.parent.mkdir(parents=True, exist_ok=True)
            commands, diags = await self.toolchain.compile(self.source_path / self.source, self.output, self.private_cxx_flags)
            self.parent.diagnostics.insert(diags, str(self.source))
        except CompilationFailure as err:
            self.parent.diagnostics.insert(err.diags, str(self.source))
            err.target = self
            raise
        self.cache['compile_args'] = [str(a) for a in commands[0]]
        self.debug('scanning dependencies of %s', self.source.name)
        deps = await self.toolchain.scan_dependencies(self.source_path / self.source, self.private_cxx_flags, self.build_path)
        deps = [d for d in deps
                if self.makefile.root.source_path in Path(d).parents
                or self.build_path in Path(d).parents]
        self.cache['deps'] = deps


class OptionSet:
    def __init__(self, parent: 'CXXTarget',
                 name: str,
                 public: list | set = set(),
                 private: list | set = set(),
                 transform: t.Callable = None) -> None:
        self._parent = parent
        self._name = name
        self._transform = transform or (lambda x: x)
        self._public = list(public)
        self._private = list(private)

    @property
    def private(self) -> list:
        return unique(self._transform(self._private))

    @property
    def public(self) -> list:
        items: list = self._transform(self._public)
        for dep in self._parent.cxx_dependencies:
            items.extend(getattr(dep, self._name).public)
        return unique(items)
    
    def __recurse_public(self) -> list:
        opts = list()
        for dep in self._parent.cxx_dependencies:
            opts.extend(getattr(dep, self._name).__recurse_public())
        opts.extend(self.public)
        return opts

    @property
    def all(self) -> list:
        opts = list()
        for dep in self._parent.cxx_dependencies:
            opts.extend(getattr(dep, self._name).__recurse_public())
        opts.extend(self.private)
        opts.extend(self.public)
        return unique(opts)

    @property
    def private_raw(self) -> list:
        return self._private

    @property
    def public_raw(self) -> list:
        return self._public

    def add(self, *values, public=False):
        if public:
            for value in values:
                if not value in self._public:
                    self._public.append(value)
        else:
            for value in values:
                if not value in self._private:
                    self._private.append(value)

    def update(self, values: 'OptionSet', private=False):
        self._public = values._public
        if private:
            self._private = values._private

class CXXTarget(Target, internal=True):
    public_includes: set[str] = set()
    private_includes: set[str] = set()

    public_compile_options: set[str] = set()
    private_compile_options: set[str] = set()
    
    public_compile_definitions: set[str] = set()
    private_compile_definitions: set[str] = set()

    public_lib_paths: set[str] = set()
    private_lib_paths: set[str] = set()

    public_link_libraries: set[str] = set()
    private_link_libraries: set[str] = set()

    public_link_options: set[str] = set()
    private_link_options: set[str] = set()

    def __init__(self,
                 *args,
                 **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.toolchain : Toolchain = self.context.get('cxx_target_toolchain')

        self.includes = OptionSet(self, 'includes',
                                #   self.public_includes, self.private_includes,
                                  transform=self.toolchain.make_include_options)

        self.compile_options = OptionSet(self, 'compile_options',
                                         self.public_compile_options, self.private_compile_options)

        self.link_libraries = OptionSet(self, 'link_libraries',
                                        self.public_link_libraries, self.private_link_libraries, transform=self.toolchain.make_link_options)
        
        self.library_paths = OptionSet(self, 'library_paths',
                                        self.public_lib_paths, self.private_lib_paths, transform=self.toolchain.make_libpath_options)

        self.compile_definitions = OptionSet(self, 'compile_definitions',
                                             self.public_compile_definitions, self.private_compile_definitions, transform=self.toolchain.make_compile_definitions)

        self.link_options = OptionSet(self, 'link_options',
                                      self.public_link_options, self.private_link_options)

        for path in self.public_includes:
            path = Path(path)
            self.includes.add(
                path if path.is_absolute() else self.source_path / path,
                public=True)

        for path in self.private_includes:
            path = Path(path)
            self.includes.add(
                path if path.is_absolute() else self.source_path / path)

    @property
    def cxx_dependencies(self) -> list['CXXTarget']:
        return [dep for dep in self.dependencies if isinstance(dep, CXXTarget)]

    @property
    def library_dependencies(self) -> list['Library']:
        return [dep for dep in self.dependencies if isinstance(dep, Library)]

    @cached_property
    def lib_paths(self) -> list[str]:
        tmp = set()
        for dep in self.cxx_dependencies:
            tmp.update(dep.lib_paths)
        tmp.update(self.library_paths.public)
        # # TODO move create private_libs()
        tmp.update(self.library_paths.private)
        return list(sorted(tmp))

    @cached_property
    def libs(self) -> LibraryList:
        tmp = LibraryList()
        for dep in reversed(self.cxx_dependencies):
            tmp.extend(dep.libs)
        tmp.extend(self.link_libraries.public)
        # # TODO move create private_libs()
        tmp.extend(self.link_libraries.private)
        return tmp

    @cached_property
    def cxx_flags(self):
        flags = self.includes.public
        flags.extend(self.compile_options.public)
        flags.extend(self.compile_definitions.public)
        for dep in self.cxx_dependencies:
            flags.extend(dep.cxx_flags)
        return unique(flags)

    @cached_property
    def private_cxx_flags(self):
        flags = self.includes.private
        flags.extend(self.cxx_flags)
        flags.extend(self.compile_options.private)
        flags.extend(self.compile_definitions.private)
        return unique(flags)

StrOrPath = str|Path
StrOrPathIterable = Iterable[StrOrPath]

class CXXObjectsTarget(CXXTarget, internal=True):
    sources: StrOrPathIterable|t.Callable[[], StrOrPathIterable] = set()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.objs: list[CXXObject] = list()

    @cache.once_method
    def _init_sources(self):
        if callable(self.sources):
            self.sources = list(self.sources())
        if not isinstance(self.sources, Iterable):
            assert callable(
                self.sources), f'{self.name} sources parameter should be an iterable or a callable returning an iterable'
        sources = list()
        source_root = Path(os.path.commonprefix(self.sources))
        for source in self.sources:
            source = Path(source)
            if source.is_absolute():
                root = source_root
            else:
                root = self.source_path
            sources.append(source)
            self.objs.append(
                CXXObject(Path(source), self, root=root))
        self.sources = sources
            

    @property
    def file_dependencies(self):
        return unique(super().file_dependencies, *[o.file_dependencies for o in self.objs])
    
    @cached_property
    def up_to_date(self):
        for obj in self.objs:
            if not obj.up_to_date:
                return False
        return super().up_to_date

    @property
    def headers(self):
        return [f for f in self.file_dependencies if f.suffix.startswith('.h')]

    async def __initialize__(self):
        self._init_sources()
        async with asyncio.TaskGroup(f'initializing {self.name}\'s objects') as group:
            for obj in self.objs:
                group.create_task(obj.initialize())
                # self.load_dependency(obj)

    async def __build__(self):
        # compile objects
        async with asyncio.TaskGroup(f'building {self.name}\'s objects') as group:
            for dep in self.objs:
                group.create_task(dep.build())

    async def __clean__(self):
        async with asyncio.TaskGroup(f'cleaning {self.name}\'s objects') as group:
            for dep in self.objs:
                group.create_task(dep.clean())
        return await super().__clean__()


class LibraryType(str, Enum):
    AUTO = 'auto'
    STATIC = 'static'
    SHARED = 'shared'
    INTERFACE = 'interface'


class Library(CXXObjectsTarget, internal=True):

    header_match = r'.+'
    library_type: LibraryType = LibraryType.AUTO

    @property
    def static(self) -> bool:
        return self.library_type == LibraryType.STATIC

    @property
    def shared(self) -> bool:
        return self.library_type == LibraryType.SHARED

    @property
    def interface(self) -> bool:
        return self.library_type == LibraryType.INTERFACE

    @cached_property
    def lib_paths(self) -> list[str]:
        tmp = super().lib_paths
        if not self.interface:
            tmp.extend(self.toolchain.make_libpath_options([self.output]))
        return list(sorted(tmp))
    
    @property
    def libs(self) -> list[str]:
        if not self.interface:
            libs = LibraryList()
            libs.extend(self.toolchain.make_link_options([self.output]))
            libs.extend(super().libs)
        else:
            libs = super().libs
        
        return libs

    async def __initialize__(self):
        self._init_sources()

        if self.library_type == LibraryType.AUTO:
            if len(self.sources) == 0:
                self.library_type = LibraryType.INTERFACE
            else:
                self.library_type = LibraryType.STATIC

        from .msvc_toolchain import MSVCToolchain
        if self.shared and isinstance(self.toolchain, MSVCToolchain):
            self.compile_definitions.add(f'{self.name.upper()}_EXPORT=1')

        if self.library_type != LibraryType.INTERFACE:
            self.output = self.toolchain.make_library_name(self.name, self.shared)
        else:
            self.output = f"lib{self.name}.stamp"
        await super().__initialize__()

        previous_args = self.cache.get('generate_args')
        generate = None
        match self.library_type:
            case LibraryType.STATIC:
                generate = self.toolchain.static_lib
            case LibraryType.SHARED:
                generate = self.toolchain.shared_lib
        if generate is not None:
            if previous_args and \
                    previous_args != await generate(
                        [obj.routput for obj in self.objs], self.output, self.libs, dry_run=True):
                self.__dirty = True
            else:
                self.__dirty = False
        else:
            self.__dirty = False

    @cached_property
    def up_to_date(self):
        if self.__dirty:
            return False
        return super().up_to_date

    async def __build__(self):
        await super().__build__()

        self.info(
            'creating %s library %s...', self.library_type.name.lower(), self.output.name)

        if self.static:
            await self.toolchain.static_lib([obj.routput for obj in self.objs], self.output, self.libs)
        elif self.shared:
            await self.toolchain.shared_lib([obj.routput for obj in self.objs], self.output, {*self.private_cxx_flags, *self.libs})
            from .msvc_toolchain import MSVCToolchain
            if isinstance(self.toolchain, MSVCToolchain):
                self.compile_definitions.add(
                    f'{self.name.upper()}_IMPORT=1', public=True)
        else:
            assert self.interface
            self.output.touch()

        self.debug('done')

    @asyncio.cached
    async def install(self, settings: InstallSettings, mode: InstallMode) -> list[Path]:
        if mode == InstallMode.user and not self.shared:
            return list()

        await self.build()

        tasks = list()

        if settings.create_pkg_config:
            from dan.pkgconfig.package import create_pkg_config
            tasks.append(create_pkg_config(self, settings))

        async def do_install(src: Path, dest: Path):
            if dest.exists() and dest.younger_than(src):
                self.info('%s is up-to-date', dest)
            else:
                self.debug('installing %s', dest)
                dest.parent.mkdir(parents=True, exist_ok=True)
                await aiofiles.copy(src, dest)
            return dest

        dest = settings.libraries_destination / self.output.name
        self.info('installing %s to %s', self.name, dest)

        if not self.interface:
            tasks.append(do_install(self.output, dest))

        if mode == InstallMode.dev:
            header_expr = re.compile(self.header_match)
            includes_dest = settings.includes_destination
            for public_include_dir in self.includes.public_raw:
                headers = public_include_dir.rglob('*.h*')
                for header in headers:
                    if header_expr.match(str(header)):
                        dest = includes_dest / \
                            header.relative_to(public_include_dir)
                        tasks.append(do_install(header, dest))

            for obj in self.objs:
                for dbg_file in self.toolchain.debug_files(obj.output):
                    tasks.append(do_install(dbg_file, settings.libraries_destination / dbg_file.name))

        tasks.insert(0, super().install(settings, mode))

        result = list()
        for tchunk in chunks(tasks, 100):
            result.append(await asyncio.gather(*tchunk))

        return result

class Module(CXXObjectsTarget, internal=True):
    def __init__(self, name: str, sources: list[str], *args, **kwargs):
        super().__init__(name, sources, *args, **kwargs)

    @property
    def cxx_flags(self):
        return {*self.toolchain.cxxmodules_flags, *super().cxx_flags}

    async def __build__(self):
        return await super().__build__()

class Executable(CXXObjectsTarget, internal=True):

    installed = True

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.output = self.toolchain.make_executable_name(self.name)
        self.__dirty = False

    async def __initialize__(self):
        await super().__initialize__()

        previous_args = self.cache.get('link_args')
        if previous_args:
            args = self.toolchain.make_link_commands([obj.routput for obj in self.objs], self.output,
                                                     [*self.lib_paths, *self.libs, *self.link_options.public, *self.link_options.private])[0]
            args = [str(a) for a in args]
            if sorted(previous_args) != sorted(args):
                self.__dirty = True

    @cached_property
    def up_to_date(self):
        if self.__dirty:
            return False
        return super().up_to_date

    async def __build__(self):
        await super().__build__()

        # link
        self.info('linking %s...', self.output.name)
        try:
            commands, diags = await self.toolchain.link([obj.routput for obj in self.objs], self.output,
                                                        [*self.lib_paths, *self.libs, *self.link_options.public, *self.link_options.private])
            self.diagnostics.insert(diags, str(self.output))
        except LinkageFailure as err:
            self.diagnostics.insert(err.diags, str(self.output))
            err.target = self
            raise
        self.cache['link_args'] = [str(a) for a in commands[0]]
        self.debug('done')

    @asyncio.cached
    async def install(self, settings: InstallSettings, mode: InstallMode) -> list[Path]:
        dest = settings.runtime_destination / self.output.name
        if dest.exists() and dest.younger_than(self.output):
            self.info('%s is up-to-date', dest)
        else:
            self.info('installing %s', dest)
            dest.parent.mkdir(parents=True, exist_ok=True)
            await aiofiles.copy(self.output, dest)
        return [dest]

    async def execute(self, *args, build=True, **kwargs):
        if build:
            await self.build()
        return await async_run([self.output, *args], logger=self, env=self.toolchain.env, **kwargs)



import re


class Version:
    version_pattern = re.compile(r'[\.-]')

    def __init__(self, *args) -> None:
        if len(args) == 1 and isinstance(args[0], str):
            parts = Version.version_pattern.split(args[0])
            args = list()
            # cleanup
            for a in parts:
                # drop non numeric parts (ie.: v0.11.0 -> 0.11.0, 1.2-preview5, mylib-3.2, etc...)
                while not a.isnumeric() and len(a):
                    a = a[1:]
                if len(a):
                    args.append(a)
        else:
            args = list(args)
        
        self._parts = tuple(int(a) for a in args)

    @property
    def major(self):
        return self._parts[0]

    @property
    def minor(self):
        if len(self._parts) > 1:
            return self._parts[1]

    @property
    def patch(self):
        if len(self._parts) > 2:
            return self._parts[2]
        
    @property
    def build(self):
        if len(self._parts) > 3:
            return self._parts[3]

    def __eq__(self, other):
        if isinstance(other, str):
            other = Version(other)
        elif not isinstance(other, Version):
            return False
        if self.major != other.major:
            return False
        if self.minor != other.minor:
            return False
        if self.patch != other.patch:
            return False
        if self.build != other.build:
            return False
        return True
    
    def is_compatible(self, other):
        if isinstance(other, str):
            other = Version(other)
        elif not isinstance(other, Version):
            return False
        if self.major != other.major:
            return False
        if self.minor and other.minor and self.minor != other.minor:
            return False
        if self.patch and other.patch and self.patch != other.patch:
            return False
        if self.build and other.build and self.build != other.build:
            return False
        return True

    def __gt__(self, other: 'Version'):
        if isinstance(other, str):
            other = Version(other)
        elif not isinstance(other, Version):
            return False
        for mine, their in zip(self._parts, other._parts):
            if mine > their:
                return True
            if mine < their:
                return False
        return True


    def __ge__(self, other: 'Version'):
        if isinstance(other, str):
            other = Version(other)
        elif not isinstance(other, Version):
            return False
        for mine, their in zip(self._parts, other._parts):
            if mine > their:
                return True
            if mine < their:
                return False
        return True
    
    def __lt__(self, other: 'Version'):
        if isinstance(other, str):
            other = Version(other)
        elif not isinstance(other, Version):
            return False
        for mine, their in zip(self._parts, other._parts):
            if mine < their:
                return True
            if mine > their:
                return False
        return False

    
    def __le__(self, other: 'Version'):
        if isinstance(other, str):
            other = Version(other)
        elif not isinstance(other, Version):
            return False
        for mine, their in zip(self._parts, other._parts):
            if mine < their:
                return True
            if mine > their:
                return False
        return True

    def __str__(self) -> str:
        res = str(self.major)
        if self.minor is not None:
            res += f'.{self.minor}'
            if self.patch is not None:
                res += f'.{self.patch}'
                if self.build is not None:
                    res += f'.{self.build}'
        return res
    
    def __repr__(self) -> str:
        return f'Version[{self}]'
    
    def __hash__(self) -> int:
        h = 0
        for part in self._parts:
            h ^= hash(part)
        return h


class VersionSpec:

    @staticmethod
    def parse(data: str) -> tuple[str|None, 'VersionSpec']:
        m = re.match(r'(.+?)?\s+?([><]=?|=)\s+([\d\.]+)', data)
        if m:
            name = m[1]
            op = m[2]
            version = Version(m[3])
            return name, VersionSpec(version, op)
        return None, None


    def __init__(self, version: Version, op: str) -> None:
        self.version = version
        self.op = op
        
    def is_compatible(self, version: Version):
        if isinstance(version, VersionSpec):
            version = version.version
        match self.op:
            case '==':
                return version == self.version
            case '=':
                return version.is_compatible(self.version)
            case '>':
                return version > self.version
            case '>=':
                return version >= self.version
            case '<':
                return version < self.version
            case '<=':
                return version <= self.version
            case _:
                return False
            
    def __str__(self) -> str:
        return f'{self.op} {self.version}'



from dataclasses import dataclass, field
from enum import Enum

from dan.core.pathlib import Path


class InstallMode(Enum):
    user = 0
    dev = 1


class BuildType(Enum):
    debug = 0
    release = 1
    release_min_size = 2
    release_debug_infos = 3

    @property
    def is_debug_mode(self):
        """Return true if the build type should produce debug symbols (ie.: debug and release_debug_infos)"""
        return self in (BuildType.debug, BuildType.release_debug_infos)

@dataclass(eq=True, unsafe_hash=True)
class InstallSettings:
    destination: str = '/usr/local'
    runtime_prefix: str = 'bin'
    libraries_prefix: str = 'lib'
    includes_prefix: str = 'include'
    data_prefix:str = 'share'
    create_pkg_config: bool = True

    @property
    def runtime_destination(self):
        return Path(self.destination).absolute() / self.runtime_prefix

    @property
    def libraries_destination(self):
        return Path(self.destination).absolute() / self.libraries_prefix

    @property
    def data_destination(self):
        return Path(self.destination).absolute() / self.data_prefix

    @property
    def includes_destination(self):
        return Path(self.destination).absolute() / self.includes_prefix

@dataclass
class ToolchainSettings:
    cxx_flags: list[str] = field(default_factory=lambda: list())

@dataclass
class Settings:
    build_type: BuildType = BuildType.debug
    install: InstallSettings = field(default_factory=lambda: InstallSettings())
    target: ToolchainSettings = field(default_factory=lambda: ToolchainSettings())


def safe_load(name: str, value,  t: type):
    if t is not None and not isinstance(value, t):
        err = f'value {name} should be of type {t}'
        if type(value) == str:
            if issubclass(t, Enum):
                names = [n.lower()
                         for n in t._member_names_]
                value = value.lower()
                if value in names:
                    value = t(names.index(value))
                else:
                    err = f'option {name} should be one of {names}'
            else:
                import json
                value = json.loads(value)
            if not isinstance(value, t):
                raise RuntimeError(err)
        else:
            raise RuntimeError(err)
    return value



from functools import cached_property
import json
import typing as t

import aiofiles
from dan.core.runners import sync_run
from dan.core.settings import BuildType
from dan.core.utils import unique
from dan.cxx.toolchain import CommandArgsList, RuntimeType, Toolchain, Path, FileDependency
from dan.core import diagnostics as diag
from dan.core.pm import re_match
import os


class MSVCToolchain(Toolchain):
    def __init__(self, data, *args, **kwargs):
        Toolchain.__init__(self, data, *args, **kwargs)
        self.cc = Path(data['cc'])
        self.cxx = self.cc
        self.lnk = Path(data['link'])
        self.lib = Path(data['lib'])
        self.env: dict[str, str] = data['env']
        if 'INCLUDE' in self.env:
            self._default_include_paths = self.env['INCLUDE'].split(os.pathsep)
        else:
            self._default_include_paths = list()
    
    @cached_property
    def common_flags(self):
        flags = [
            '/nologo',
        ]
        if self.build_type.is_debug_mode:
            flags.append('/DEBUG')
        return flags

    @cached_property
    def default_cflags(self):
        flags = [
            '/EHsc',
            '/GA',
        ]
        rt = '/MD' if self.runtime == RuntimeType.dynamic else '/MT'
        match self.build_type:
            case BuildType.debug:
                flags.extend([
                    f'{rt}d',
                    '/Gd',
                    '/ZI',
                    '/FS',
                ])
            case BuildType.release:
                flags.extend((rt, '/O2', '/DNDEBUG'))
            case BuildType.release_min_size:
                flags.extend((rt, '/Os', '/DNDEBUG'))
            case  BuildType.release_debug_infos:
                flags.extend((rt, '/O2', '/DNDEBUG'))

        return flags

    @property
    def default_cxxflags(self):
        return [f'/std:c++{self.cpp_std}', *self.settings.cxx_flags]
    
    async def get_default_include_paths(self, lang = 'c++') -> list[str]:
        return self._default_include_paths

    def has_cxx_compile_options(self, *opts) -> bool:
        _, err, _ = sync_run([self.cxx, *opts], no_raise=True)
        # D9002 => unknown option
        return err.splitlines()[0].find('D9002') == 0

    def make_include_options(self, include_paths: set[Path]) -> list[str]:
        return [f'/I{p}' for p in include_paths]

    def make_libpath_options(self, libraries: set[Path | str]) -> list[str]:
        lib_paths = list()
        for lib in libraries:
            if isinstance(lib, Path):
                lib_paths.append(f'/LIBPATH:{lib.parent}')
        return lib_paths

    def make_link_options(self, libraries: set[Path | str]) -> list[str]:
        libs = list()
        for lib in libraries:
            if isinstance(lib, Path):
                libs.append(f'{lib.stem}.lib')
            else:
                assert isinstance(lib, str)
                libs.append(f'{lib}.lib')
        return libs

    def make_compile_definitions(self, definitions: set[str]) -> list[str]:
        return [f'/D{d}' for d in definitions]

    def make_library_name(self, basename: str, shared: bool) -> str:
        return f'{basename}.{"dll" if shared else "lib"}'

    def make_executable_name(self, basename: str) -> str:
        return f'{basename}.exe'

    def from_unix_flags(self, flags: list[str]):
        out = list()
        for flag in flags:
            match re_match(flag):
                case r'-L(.+)' as m:
                    out.append(f'/LIBPATH:{m[1]}')
                case r'-l(.+)' as m:
                    out.append(f'{m[1]}.lib')
                case r'-I(.+)' as m:
                    out.append(f'/I{m[1]}')
                case _:
                    out.append(flag)
        return out

    def to_unix_flags(self, flags: list[str]):
        out = list()
        for flag in flags:
            flag = flag.replace('"', '')
            match re_match(flag):
                case r'/LIBPATH:(.+)' as m:
                    out.append(f'-L{m[1]}')
                case r'(.+).lib' as m:
                    out.append(f'-l{m[1]}')
                case r'/I(.+)' as m:
                    out.append(f'-I{m[1]}')
                case _:
                    out.append(flag)
        return out

    async def scan_dependencies(self, sourcefile: Path, options: list[str], build_path: Path) -> set[FileDependency]:
        deps_path = build_path / sourcefile.with_suffix(".json").name
        deps = set()
        if deps_path.exists():
            async with aiofiles.open(deps_path, 'r') as f:
                content = await f.read()
                data = json.loads(content)
                deps = set(data['Data']['Includes'])
        return deps

    def compile_generated_files(self, output: Path) -> set[Path]:
        return {}

    def debug_files(self, output: Path) -> set[Path]:
        if not self.build_type.is_debug_mode:
            return {}
        return {output.with_suffix('.pdb')}

    @property
    def cxxmodules_flags(self) -> list[str]:
        return list()

    def make_compile_commands(self, sourcefile: Path, output: Path, options: list[str]) -> CommandArgsList:
        deps = output.parent / sourcefile.with_suffix(".json").name
        args = [self.cc, *unique(self.common_flags, self.default_cflags, self.default_cxxflags, options),
                '/sourceDependencies', deps,
                f'/Fo{str(output)}', '/c', str(sourcefile)]
        if self.build_type.is_debug_mode:
            args.append(f'/Fd{str(output.with_suffix(".pdb"))}')
        return [args]

    def make_link_commands(self, objects: set[Path], output: Path, options: list[str]) -> CommandArgsList:
        return [[self.lnk, *self.common_flags, *options, *objects, f'/OUT:{str(output)}']]

    def make_static_lib_commands(self, objects: set[Path], output: Path, options: list[str]) -> CommandArgsList:
        return [[self.lib, *self.common_flags, *objects, f'/OUT:{output}']]

    def make_shared_lib_commands(self, objects: set[Path], output: Path, options: list[str]) -> CommandArgsList:
        return [[self.lnk, *self.common_flags,
                f'/IMPLIB:{output.with_suffix(".lib")}', '/DLL', *options, *objects, f'/OUT:{output.with_suffix(".dll")}']]

    async def _handle_compile_output(self, lines) -> t.Iterable[diag.Diagnostic]:
        async for line in lines:
            line = line.strip()
            match re_match(line):
                case r'(.+)\((\d+)\):\s+(?:fatal\s+)?(error|warning)\s(\w+\d+):\s(.+)$' as m:
                    yield diag.Diagnostic(
                        message=m[5].strip(),
                        range=diag.Range(start=diag.Position(line=int(m[2])-1)),
                        code=m[4],
                        severity=diag.Severity[m[3].upper()],
                        source=self.type,
                        filename=m[1])
                case _:
                    self._logger.debug('Unhandled line: %s', line)

    async def _handle_link_output(self, lines) -> t.Iterable[diag.Diagnostic]:
        async for line in lines:
            match re_match(line):
                case r'(.+?)\s?:\serror\s(\w+\d+):\s(.+)$' as m:
                    yield diag.Diagnostic(message=m[3].strip(), code=m[2], source=self.type, filename=m[1])
                case r'LINK\s?:\s?fatal\s+error\s+(\w+\d+):\s+(.+)$' as m:
                    yield diag.Diagnostic(message=m[2].strip(), code=m[1], source=self.type)
                case _:
                    self._logger.debug('Unhandled line: %s', line)



import asyncio
from aiofiles import *
from aiofiles import os

import os as sync_os
import stat

from dan.core.pathlib import Path

async def _rmtree(path: Path):
    for root, dirs, files in sync_os.walk(path, topdown=False):
        clean_files = [os.remove(
            sync_os.path.join(root, name)) for name in files]
        clean_dirs = [os.rmdir(
            sync_os.path.join(root, name)) for name in dirs]
        await asyncio.gather(*clean_files)
        await asyncio.gather(*clean_dirs)
    await os.rmdir(path)

async def _remove_force(path: Path):
    sync_os.chmod(path, stat.S_IWRITE)
    await os.remove(path)

async def _rmtree_force(path: Path):
    for root, dirs, files in sync_os.walk(path, topdown=False):
        clean_files = [_remove_force(
            Path(root) / name) for name in files]
        clean_dirs = [os.rmdir(
            sync_os.path.join(root, name)) for name in dirs]
        await asyncio.gather(*clean_files)
        await asyncio.gather(*clean_dirs)
    await os.rmdir(path)

async def rmtree(path: Path, force=False):
    if force:
        await _rmtree_force(path)
    else:
        await _rmtree(path)

async def copy(src : Path, dest : Path, chunk_size=2048):
    if dest.is_dir():
        dest = dest / src.name
    dest.parent.mkdir(exist_ok=True, parents=True)
    async with open(src, 'rb') as s,\
               open(dest, 'wb') as d:
        while True:
            chunk = await s.read(chunk_size)
            if len(chunk) == 0:
                break
            await d.write(chunk)
    dest.chmod(src.stat().st_mode)



import os
from click import Path
import tqdm
from dan.core import aiofiles, asyncio
from dan.core.target import Target
import aiohttp
import tarfile
import zipfile


async def fetch_file(url, dest: Path):
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as resp:
            if resp.status != 200:
                message = resp.read()
                raise RuntimeError(f'unable to fetch {url}: {message.decode()}')
            size = int(resp.headers.get('content-length', 0))

            with tqdm.tqdm(
                desc=f'downloading {dest.name}', total=size // 1024, leave=False, unit='Ko'
            ) as progressbar:
                async with aiofiles.open(dest, mode='wb') as f:
                    async for chunk in resp.content.iter_chunked(1024):
                        await f.write(chunk)
                        progressbar.update(len(chunk) // 1024)


class TarSources(Target, internal=True):

    url: str

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.output: Path = 'sources'

    async def __build__(self):
        self.info(f'downloading {self.url}')
        archive_name = self.url.split("/")[-1]
        await fetch_file(self.url, self.build_path / archive_name)
        self.info(f'extracting {archive_name}')
        if archive_name.endswith('.zip'):
            with zipfile.ZipFile(self.build_path / archive_name) as f:
                root = os.path.commonprefix(f.namelist())
                f.extractall(self.output.parent)
        else:
            with tarfile.open(self.build_path / archive_name) as f:
                root = os.path.commonprefix(f.getnames())
                f.extractall(self.output.parent)
        async with asyncio.TaskGroup() as g:
            g.create_task(aiofiles.os.rename(self.output.parent / root, self.output))
            g.create_task(aiofiles.os.remove(self.build_path / archive_name))



import os
import sys
from dan import logging
import asyncio

from dan.core.find import find_file
from dan.core.pathlib import Path
from dan.cli import click

from dan.core import diagnostics
from dan.core.cache import Cache
from dan.core.settings import Settings
from dan.cxx.targets import Executable


from dan.make import InstallMode, Make
from dan.cli.vscode import Code


_minimal_options = [
    click.option('--build-path', '-B', help='Path where dan has been initialized.',
                 type=click.Path(resolve_path=True, path_type=Path), required=True, default='build', envvar='DAN_BUILD_PATH'),
]

_common_opts = [
    *_minimal_options,
    click.option('--quiet', '-q', is_flag=True,
                 help='Dont print informations (errors only).', envvar='DAN_QUIET'),
    click.option('--verbose', '-v', is_flag=True,
                 help='Pring debug informations.', envvar='DAN_VERBOSE'),
    click.option('--jobs', '-j',
                 help='Maximum jobs.', default=None, type=int, envvar='DAN_JOBS'),
    click.option('--no-progress', is_flag=True,
                 help='Disable progress bars', envvar='DAN_NOPROGRESS'),
]
_base_help_ = '''
  PATH          Either build or source directory.
  [TARGETS...]  Targets to process.
'''


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func
    return _add_options


common_opts = add_options(_common_opts)
minimal_options = add_options(_minimal_options)


class CommandsContext:
    def __init__(self, *args, **kwds) -> None:
        self._make_args = [*args]
        self._make_kwds = {**kwds}
        self._make = None

    def __call__(self, *args, **kwds):
        if len(args):
            self._make_args.extend(*args)
        self._make_kwds.update(**kwds)

    @property
    def make(self):
        if self._make is None:
            self._make = Make(*self._make_args, **self._make_kwds)
        return self._make


pass_context = click.make_pass_decorator(CommandsContext)

@pass_context
def show_diags(ctx: CommandsContext):
    if diagnostics.enabled:
        diags = ctx.make.diagnostics
        if diags:
            click.echo(f'DIAGNOSTICS: {diags.to_json()}')

@click.group()
@click.version_option(package_name='dan-build')
@click.option('--quiet', '-q', is_flag=True,
              help='Dont print informations (errors only)')
@click.option('--verbose', '-v', is_flag=True,
              help='Pring debug informations')
@click.option('--jobs', '-j',
              help='Maximum jobs', default=None, type=int)
@click.pass_context
def cli(ctx: click.AsyncContext, **kwds):
    ctx.obj = CommandsContext(**kwds)
    ctx.call_on_close(show_diags)


@cli.command()
@click.option('--verbose', '-v', is_flag=True,
              help='Pring debug informations')
@click.option('--toolchain', '-t', help='The toolchain to use',
              type=click.ToolchainParamType())
@click.option('--setting', '-s', 'settings', help='Set or change a setting', multiple=True, type=click.SettingsParamType(Settings))
@click.option('--option', '-o', 'options', help='Set or change an option', multiple=True, type=click.OptionsParamType())
@click.option('--build-path', '-B', help='Path where dan has been initialized.',
              type=click.Path(resolve_path=True, path_type=Path), required=True, default='build', envvar='DAN_BUILD_PATH')
@click.option('--source-path', '-S', help='Path where source is located.',
              type=click.Path(resolve_path=True, path_type=Path), required=True, default='.')
@pass_context
async def configure(ctx: CommandsContext, toolchain: str, settings: tuple[str], options: tuple[str], source_path: Path, **kwds):
    """Configure dan project"""
    ctx(**kwds)  # update kwds
    if toolchain is None and ctx.make.config.toolchain is None:
        toolchain = click.prompt('Toolchain', type=click.ToolchainParamType(), default='default')

    await ctx.make.configure(source_path, toolchain)

    if len(settings):
        await ctx.make.apply_settings(*settings)

    # NOTE: intializing make after applying setting
    #       to check settings are valid implicitly (cache save skipped)
    await ctx.make.initialize()

    if len(options):
        await ctx.make.apply_options(*options)


@cli.command()
@click.option('--for-install', is_flag=True, help='Build for install purpose (will update rpaths [posix only])')
@common_opts
@click.option('--force', '-f', is_flag=True,
              help='Clean before building')
@click.argument('TARGETS', nargs=-1, type=click.TargetParamType())
@pass_context
async def build(ctx: CommandsContext, force=False, **kwds):
    """Build targets"""
    ctx(**kwds)  # update kwds
    if force:
        await ctx.make.clean()
    await ctx.make.build()
    # from dan.cxx import target_toolchain
    # target_toolchain.compile_commands.update()


@cli.command()
@common_opts
@click.argument('MODE', type=click.Choice([v.name for v in InstallMode]), default=InstallMode.user.name)
@click.argument('TARGETS', nargs=-1, type=click.TargetParamType())
@pass_context
async def install(ctx: CommandsContext, mode: str, **kwargs):
    """Install targets"""
    ctx(**kwargs)
    mode = InstallMode[mode]
    await ctx.make.install(mode)


@cli.command()
@click.option('--verbose', '-v', is_flag=True,
              help='Pring debug informations')
@click.option('--yes', '-y', is_flag=True, help='Proceed without asking')
@click.option('--root', '-r', help='Root path to search for installation manifest', type=click.Path(exists=True, file_okay=False))
@click.argument('NAME')
def uninstall(verbose: bool, yes: bool, root: str, name: str):
    """Uninstall previous installation"""
    logging.getLogger().setLevel(logging.DEBUG if verbose else logging.INFO)
    if root:
        paths = [root]
    else:
        paths = [
            '~/.local/share/dan',
            '/usr/local/share/dan',
            '/usr/share/dan',
        ]
    manifest = find_file(f'{name}-manifest.txt$', paths=paths)
    with open(manifest, 'r') as f:
        files = [(manifest.parent / mf.strip()).resolve()
                 for mf in f.readlines()]
    to_be_removed = '\n'.join([f" - {f}" for f in files])
    yes = yes or click.confirm(
        f'Following files will be removed:\n {to_be_removed}\nProceed ?')
    if yes:
        def rm_empty(dir: Path):
            if dir.is_empty:
                click.logger.debug(f'removing empty directory: {dir}')
                os.rmdir(dir)
                rm_empty(dir.parent)

        for f in files:
            click.logger.debug(f'removing: {f}')
            os.remove(f)
            rm_empty(f.parent)

        os.remove(manifest)
        rm_empty(manifest.parent)

@cli.group()
@pass_context
def ls(ctx: CommandsContext):
    """Inspect stuff"""
    pass

@ls.command()
@click.option('-a', '--all', 'all', is_flag=True, help='Show all targets (not only defaulted ones)')
@click.option('-t', '--type', 'show_type', is_flag=True, help='Show target\'s type')
@common_opts
@click.argument('TARGETS', nargs=-1)
@pass_context
async def targets(ctx: CommandsContext, all: bool, show_type: bool, **kwargs):
    """List targets"""
    kwargs['quiet'] = True
    ctx(**kwargs)
    await ctx.make.initialize()
    out = []
    for target in ctx.make.targets:
        if show_type:
            out.append(target.fullname + ' - ' + type(target).__name__)
        else:
            out.append(target.fullname)
    click.echo('\n'.join(out))


@ls.command()
@common_opts
@click.argument('TARGETS', nargs=-1)
@pass_context
async def tests(ctx: CommandsContext, **kwargs):
    """List tests"""
    kwargs['quiet'] = True
    ctx(**kwargs)
    await ctx.make.initialize()
    for t in ctx.make.tests:
        if len(t) > 1:
            for c in t.cases:
                click.echo(f'{t.fullname}:{c.name}')
        else:
            click.echo(t.fullname)

@ls.command()
@common_opts
@click.argument('TARGETS', nargs=-1)
@pass_context
async def options(ctx: CommandsContext, **kwargs):
    """List tests"""
    kwargs['quiet'] = True
    ctx(**kwargs)
    await ctx.make.initialize()
    for o in ctx.make.all_options:
        current = ''
        if o.value != o.default:
            current = f', current: {o.value}'
        click.echo(f'{o.fullname}: {o.help} (type: {o.type.__name__}, default: {o.default}{current})')

@ls.command()
def toolchains(**kwargs):
    """List toolchains"""
    kwargs['quiet'] = True
    for name, _ in Make.toolchains()['toolchains'].items():
        click.echo(name)


@cli.command()
@common_opts
@click.argument('TARGETS', nargs=-1, type=click.TargetParamType())
@pass_context
async def clean(ctx, **kwargs):
    """Clean generated stuff"""
    ctx(**kwargs)
    await ctx.make.clean()


@cli.command()
@common_opts
@click.argument('TARGETS', nargs=-1, type=click.TargetParamType())
@pass_context
async def run(ctx, **kwargs):
    """Run executable(s)"""
    ctx(**kwargs)
    rc = await ctx.make.run()
    sys.exit(rc)


@cli.command()
@common_opts
@click.argument('TARGETS', nargs=-1, type=click.TargetParamType())
@pass_context
async def test(ctx, **kwargs):
    """Run tests"""
    ctx(**kwargs)
    rc = await ctx.make.test()
    sys.exit(rc)


@cli.command()
@click.option('-s', '--script',
              help='Use a source script to resolve compilation environment')
@click.option('-p', '--path', 'paths',
              help='Use given path for compilers lookup', multiple=True, type=click.Path(exists=True, file_okay=False))
@click.option('--verbose', '-v', is_flag=True,
              help='Pring debug informations.', envvar='DAN_VERBOSE')
def scan_toolchains(script: str, paths: list[str], verbose, **kwargs):
    """Scan system toolchains"""
    logging.getLogger().setLevel(logging.DEBUG if verbose else logging.INFO)
    from dan.cxx.detect import create_toolchains, load_env_toolchain
    if script:
        load_env_toolchain(script)
    else:
        create_toolchains(paths if len(paths) else None)


@cli.group()
def code():
    """VS-Code specific commands"""
    pass


@code.command()
@common_opts
@click.argument('TARGETS', nargs=-1)
@pass_context
async def get_targets(ctx: CommandsContext, **kwargs):
    kwargs.update({'quiet': True, 'diags': True})
    ctx(**kwargs)
    await ctx.make.initialize()
    out = []
    targets = ctx.make.context.root.all_targets
    for target in targets:
        out.append({
            'name': target.name,
            'fullname': target.fullname,
            'buildPath': str(target.build_path),
            'output': str(target.output),
            'executable': isinstance(target, Executable),
            'type': type(target).__name__
        })
    import json
    click.echo(json.dumps(out))

@code.command()
@common_opts
@click.argument('TARGETS', nargs=-1)
@pass_context
async def get_tests(ctx: CommandsContext, **kwargs):
    kwargs.update({'quiet': True, 'diags': True})
    ctx(**kwargs)
    await ctx.make.initialize()
    import json
    out = list()
    for t in ctx.make.context.root.all_tests:
        out.append(t.fullname)
        if len(t) > 1:
            for c in t.cases:
                out.append(f'{t.fullname}:{c.name}')
    click.echo(json.dumps(out))


@code.command()
@common_opts
@click.option('--pretty', is_flag=True)
@click.argument('TARGETS', nargs=-1)
@pass_context
async def get_test_suites(ctx: CommandsContext, pretty, **kwargs):
    kwargs.update({'quiet': True, 'diags': True})
    ctx(**kwargs)
    await ctx.make.initialize()
    code = Code(ctx.make)
    click.echo(code.get_test_suites(pretty))


@code.command()
def get_toolchains(**kwargs):
    import json
    click.echo(json.dumps(list(Make.toolchains()['toolchains'].keys())))


@code.command()
@click.option('--for-install', is_flag=True, help='Build for install purpose (will update rpaths [posix only])')
@common_opts
@click.option('--force', '-f', is_flag=True,
              help='Clean before building')
@click.argument('TARGETS', nargs=-1)
@pass_context
async def build(ctx: CommandsContext, force=False, **kwds):
    """Build targets (vscode version)"""
    ctx(**kwds, diags=True)  # update kwds
    if force:
        await ctx.make.clean()
    await ctx.make.build()


@code.command()
@minimal_options
@click.argument('SOURCES', nargs=-1)
@pass_context
async def get_source_configuration(ctx: CommandsContext, sources, **kwargs):
    kwargs.update({'quiet': True, 'diags': True})
    ctx(**kwargs)
    await ctx.make.initialize()
    code = Code(ctx.make)
    click.echo(await code.get_sources_configuration(sources))


@code.command()
@minimal_options
@pass_context
async def get_workspace_browse_configuration(ctx: CommandsContext, **kwargs):
    kwargs.update({'quiet': True, 'diags': True})
    ctx(**kwargs)
    await ctx.make.initialize()
    code = Code(ctx.make)
    click.echo(await code.get_workspace_browse_configuration())

@cli.result_callback()
@pass_context
async def process_result(ctx, result, **kwargs):
    await Cache.save_all()


def main():
    import sys
    try:
        cli(auto_envvar_prefix='DAN')
    except Exception as err:
        click.logger.error(str(err))
        _ex_type, _ex, tb = sys.exc_info()
        import traceback
        click.logger.debug(' '.join(traceback.format_tb(tb)))
        try:
            # wait asyncio loop to terminate
            asyncio.get_running_loop().run_until_complete()
        except Exception:
            pass
        asyncio.run(Cache.save_all())
        return -1
